# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict, List, Optional, Tuple

import mxnet as mx
import numpy as np

from gluonts_customized.core.component import Type, validated
from gluonts_customized.itertools import prod
from gluonts_customized.mx.model.deepar import DeepAREstimator
from gluonts_customized.mx.model.deepar._network import DeepARPredictionNetwork
from gluonts_customized.mx.model.deepvar_hierarchical._estimator import projection_mat
from gluonts_customized.mx.model.deepvar_hierarchical._network import coherency_error
from gluonts_customized.mx.distribution import Distribution, EmpiricalDistribution
from gluonts_customized.mx import Tensor
from gluonts_customized.mx.distribution import TransformedPiecewiseLinear

from gluonts_customized.nursery.temporal_hierarchical_forecasting.utils import utils
from gluonts_customized.nursery.temporal_hierarchical_forecasting.model.cop_deepar import (
    gluonts_fixes,
    gnn,
)


def reconcile_samples(
    reconciliation_mat: Tensor,
    samples: Tensor,
    non_negative: bool = False,
    num_iters: int = 10,
) -> Tensor:
    if not non_negative:
        return mx.nd.dot(samples, reconciliation_mat, transpose_b=True)
    else:
        # Dykstra's projection method: Projection onto the intersection of convex sets.
        x = samples
        p = mx.nd.zeros_like(x)
        q = mx.nd.zeros_like(x)
        for _ in range(num_iters):
            # Projection onto the non-negative orthant.
            y = mx.nd.relu(x + p)
            p = x + p - y

            # Projection onto the null space.
            x = mx.nd.dot(y + q, reconciliation_mat, transpose_b=True)
            q = y + q - x

        return x


class COPNetwork(mx.gluon.HybridBlock):
    @validated()
    def __init__(
        self,
        estimators: List[DeepAREstimator],
        prediction_length: int,
        temporal_hierarchy: utils.TemporalHierarchy,
        do_reconciliation: bool,
        dtype: Type,
        use_gnn: bool,
        use_mlp: bool,
        adj_mat_option: str,
        non_negative: bool = False,
        naive_reconciliation: bool = False,
        prediction: bool = False,
        loss_function: str = "crps_univariate",
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.prediction_length = prediction_length
        self.temporal_hierarchy = temporal_hierarchy
        self.use_gnn = use_gnn
        self.use_mlp = use_mlp
        self.adj_mat_option = adj_mat_option
        self.do_reconciliation = do_reconciliation
        self.non_negative = non_negative
        self.loss_function = loss_function
        self.dtype = dtype

        if naive_reconciliation:
            M = utils.naive_reconcilation_mat(
                self.temporal_hierarchy.agg_mat, self.temporal_hierarchy.nodes
            )
        else:
            M = projection_mat(S=self.temporal_hierarchy.agg_mat)
        self.M = mx.nd.array(M)

        self.estimators = estimators

        self.models = []
        with self.name_scope():
            for estimator in estimators:
                if not prediction:
                    self.network = estimator.create_training_network()
                else:
                    self.network = gluonts_fixes.create_prediction_network(
                        estimator
                    )

                self.register_child(self.network)
                self.models.append(self.network)

            if self.use_gnn:
                # GNN Layer: Do message passing for `L-1` times, where `L` is the number of levels of the hierarchy.
                self.gnn = gnn.GNN(
                    units=self.estimators[0].num_cells,
                    num_layers=len(self.temporal_hierarchy.agg_multiples) - 1,
                    adj_matrix=mx.nd.array(
                        self.temporal_hierarchy.adj_mat(
                            option=self.adj_mat_option
                        )
                    ),
                    use_mlp=self.use_mlp,
                )

    def get_target_related_feat_at_agg_level(
        self,
        agg_level: int,
        past_target: Tensor,
        past_observed_values: Tensor,
        past_is_pad: Tensor,
        future_target: Optional[Tensor] = None,
        future_observed_values: Optional[Tensor] = None,
    ) -> Dict:
        """
        Aggregate target at the given aggregate level along with updating
        observed value and pad indicators.

        :param agg_level:
        :param past_target:
        :param past_observed_values:
        :param past_is_pad:
        :param future_target:
        :param future_observed_values:
        :return:
        """
        agg_multiple = self.temporal_hierarchy.agg_multiples[agg_level]

        # Truncating the history length of the base time series to the nearest multiple.
        base_history_length = (
            past_target.shape[1] // agg_multiple
        ) * agg_multiple

        past_target_agg = (
            utils.agg_series(
                past_target.slice_axis(
                    axis=1, begin=-base_history_length, end=None
                ),
                agg_multiple=agg_multiple,
            )
            .squeeze(axis=-1)
            .slice_axis(
                axis=1, begin=-self.models[agg_level].history_length, end=None
            )
        )

        past_is_pad_agg = (
            utils.agg_series(
                past_is_pad.slice_axis(
                    axis=1, begin=-base_history_length, end=None
                ),
                agg_multiple=agg_multiple,
            )
            .squeeze(axis=-1)
            .slice_axis(
                axis=1, begin=-self.models[agg_level].history_length, end=None
            )
        )

        past_is_pad_agg = mx.nd.where(
            past_is_pad_agg == 0.0,
            mx.nd.zeros_like(past_is_pad_agg),
            mx.nd.ones_like(past_is_pad_agg),
        )

        past_observed_values_agg = (
            utils.agg_series(
                past_observed_values.slice_axis(
                    axis=1, begin=-base_history_length, end=None
                ),
                agg_multiple=agg_multiple,
            )
            .squeeze(axis=-1)
            .slice_axis(
                axis=1, begin=-self.models[agg_level].history_length, end=None
            )
        )

        past_observed_values_agg = mx.nd.where(
            # We sum observed values of base time series at `agg_multiple` time steps;
            # if all of them are 1, then the observed value for the aggregated time series is 1 and 0 otherwise.
            # We could redefine agg_series to actually compute mean, but overloading that term might cause other
            # problems later.
            past_observed_values_agg == agg_multiple,
            mx.nd.ones_like(past_observed_values_agg),
            mx.nd.zeros_like(past_observed_values_agg),
        )

        target_related_feat_agg = {
            "past_target": past_target_agg,
            "past_is_pad": past_is_pad_agg,
            "past_observed_values": past_observed_values_agg,
        }
        if future_target is not None:
            future_target_agg = utils.agg_series(
                future_target, agg_multiple=agg_multiple
            ).squeeze(axis=-1)

            future_observed_values_agg = utils.agg_series(
                future_observed_values, agg_multiple=agg_multiple
            ).squeeze(axis=-1)

            future_observed_values_agg = mx.nd.where(
                future_observed_values_agg == agg_multiple,
                mx.nd.ones_like(future_observed_values_agg),
                mx.nd.zeros_like(future_observed_values_agg),
            )

            target_related_feat_agg.update(
                {
                    "future_target": future_target_agg,
                    "future_observed_values": future_observed_values_agg,
                }
            )

        return target_related_feat_agg

    def _embeddings_to_distr(
        self,
        F,
        embeddings_at_all_levels: Tensor,
        scales: List,
    ) -> Distribution:
        distr_output = self.models[0].distr_output
        distr_args_at_all_levels: Dict = {
            arg_name: [] for arg_name in distr_output.args_dim.keys()
        }
        scales_ls = []

        start_ix = 0
        for i, num_nodes in enumerate(
            self.temporal_hierarchy.num_nodes_per_level
        ):
            end_ix = start_ix + num_nodes
            distr_args = self.models[i].proj_distr_args(
                embeddings_at_all_levels[..., start_ix:end_ix, :]
            )

            for j, arg_ls in enumerate(distr_args_at_all_levels.values()):
                arg_ls.append(distr_args[j])

            scales_ls.append(scales[i].broadcast_like(distr_args[0]))

            start_ix = end_ix

        # Last dimension contains parameters at all time-levels and aggregation can be done on it.
        distr_args_at_all_levels = {
            arg_name: F.concat(*arg_ls, dim=-1)
            for arg_name, arg_ls in distr_args_at_all_levels.items()
        }

        scale_at_all_levels = F.concat(*scales_ls, dim=-1)

        distr_at_all_levels = distr_output.distribution(
            distr_args=distr_args_at_all_levels.values(),
            scale=scale_at_all_levels,
        )

        if isinstance(distr_at_all_levels, TransformedPiecewiseLinear):
            distr_at_all_levels = TransformedPiecewiseLinear(
                base_distribution=gluonts_fixes.PiecewiseLinearWithSampling(
                    gamma=distr_at_all_levels.base_distribution.gamma,
                    slopes=distr_at_all_levels.base_distribution.slopes,
                    knot_spacings=distr_at_all_levels.base_distribution.knot_spacings,
                ),
                transforms=distr_at_all_levels.transforms,
            )

        return distr_at_all_levels

    def _distr_to_samples(
        self,
        distr_at_all_levels: Distribution,
        num_samples: int,
    ):
        if num_samples == 1:
            samples_at_all_levels = distr_at_all_levels.sample(
                num_samples=num_samples, dtype=self.dtype
            )

            # get rid of the redundant axis introduced by `sample`.
            samples_at_all_levels = samples_at_all_levels.squeeze(axis=0)
        else:
            samples_at_all_levels = distr_at_all_levels.sample_rep(
                num_samples=num_samples, dtype=self.dtype
            )

        return samples_at_all_levels


class COPDeepARTrainingNetwork(COPNetwork):
    @validated()
    def __init__(
        self,
        num_batches_per_epoch: int,
        epochs: int,
        warmstart_epoch_frac: float,
        num_samples_for_loss: int = 200,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.warmstart_epoch_frac = warmstart_epoch_frac
        self.epochs = epochs
        self.num_batches_per_epoch = num_batches_per_epoch
        self.batch_no = 0
        self.num_samples_for_loss = num_samples_for_loss

    # noinspection PyMethodOverriding,PyPep8Naming
    def hybrid_forward(
        self,
        F,
        feat_static_cat: Tensor,
        feat_static_real: Tensor,
        past_time_feat: Tensor,
        past_target: Tensor,
        past_observed_values: Tensor,
        past_is_pad: Optional[Tensor],
        future_time_feat: Tensor,
        future_target: Tensor,
        future_observed_values: Tensor,
        agg_features_dict: Dict,
    ) -> Tensor:
        """
        Computes the loss for training COPDeepAR, all inputs tensors
        representing time series have NTC layout.

        Parameters
        ----------
        F
        feat_static_cat : (batch_size, num_features)
        feat_static_real : (batch_size, num_features)
        past_time_feat : (batch_size, history_length, num_features)
        past_target : (batch_size, history_length, *target_shape)
        past_observed_values : (batch_size, history_length, *target_shape, seq_len)
        future_time_feat : (batch_size, prediction_length, num_features)
        future_target : (batch_size, prediction_length, *target_shape)
        future_observed_values : (batch_size, prediction_length, *target_shape)
        agg_features_dict: Dictionary of features for aggregated levels

        Returns loss with shape (batch_size, context + prediction_length, 1)
        -------
        """

        embeddings_at_all_levels_ls = []
        target_at_all_levels_ls = []
        scale_ls = []

        for i, agg_multiple in enumerate(
            self.temporal_hierarchy.agg_multiples
        ):
            if agg_multiple != 1:
                past_time_feat_agg = agg_features_dict[f"level_{i}"][
                    "past_time_feat_agg"
                ]
                future_time_feat_agg = agg_features_dict[f"level_{i}"][
                    "future_time_feat_agg"
                ]
            else:
                past_time_feat_agg = past_time_feat
                future_time_feat_agg = future_time_feat

            target_related_feat_agg = (
                self.get_target_related_feat_at_agg_level(
                    agg_level=i,
                    past_target=past_target,
                    past_is_pad=past_is_pad,
                    past_observed_values=past_observed_values,
                    future_target=future_target,
                    future_observed_values=future_observed_values,
                )
            )

            rnn_outputs, _, scale, _, _ = self.models[i].unroll_encoder(
                F=F,
                feat_static_cat=feat_static_cat,
                feat_static_real=feat_static_real,
                past_time_feat=past_time_feat_agg,
                future_time_feat=future_time_feat_agg,
                **target_related_feat_agg,
            )
            scale_ls.append(scale.expand_dims(axis=-1))

            # put together target sequence
            # (batch_size, seq_len, *target_shape)
            target = F.concat(
                target_related_feat_agg["past_target"].slice_axis(
                    axis=1,
                    begin=self.models[i].history_length
                    - self.models[i].context_length,
                    end=None,
                ),
                target_related_feat_agg["future_target"],
                dim=1,
            )

            # We reconcile blocks/windows of time steps: e.g., if we have 28 values of daily data, then we
            # reconcile 4 windows where each window has a length of 7 if number of leaves in the hierarchy is 7.
            window_size = self.temporal_hierarchy.num_leaves // agg_multiple
            num_windows = (
                self.models[i].context_length
                + self.models[i].prediction_length
            ) // window_size

            embeddings_at_all_levels_ls.append(
                rnn_outputs.reshape(
                    (
                        rnn_outputs.shape[0],
                        num_windows,
                        -1,
                        rnn_outputs.shape[-1],
                    )
                )
            )

            target_at_all_levels_ls.append(
                target.reshape((target.shape[0], num_windows, -1))
            )

        # Last dimension contains embeddings at all time-levels and message passing/aggregation can be done on it.
        # Shape: (bs, num_windows, total_num_time_steps_of_hierarchy, embedding_dim)
        embeddings_at_all_levels = F.concat(
            *embeddings_at_all_levels_ls, dim=-2
        )

        if self.use_gnn:
            embeddings_at_all_levels = self.gnn(embeddings_at_all_levels)

        distr_at_all_levels = self._embeddings_to_distr(
            F,
            embeddings_at_all_levels,
            scale_ls,
        )

        target_at_all_levels = F.concat(*target_at_all_levels_ls, dim=-1)

        if self.loss_function == "nll":
            loss = distr_at_all_levels.loss(x=target_at_all_levels)

            # Determine which epoch we are currently in.
            self.batch_no += 1
            epoch_no = self.batch_no // self.num_batches_per_epoch + 1
            epoch_frac = epoch_no / self.epochs

            if epoch_frac > self.warmstart_epoch_frac:
                print(
                    f"epoch_frac: {epoch_frac}. Switching the loss function to CRPS"
                )
                self.loss_function = "crps_univariate"

        else:
            samples_at_all_levels = self._distr_to_samples(
                distr_at_all_levels,
                num_samples=self.num_samples_for_loss,
            )

            if self.do_reconciliation:
                reconciled_samples_at_all_levels = reconcile_samples(
                    reconciliation_mat=self.M,
                    samples=samples_at_all_levels,
                    non_negative=self.non_negative,
                )
            else:
                reconciled_samples_at_all_levels = samples_at_all_levels

            loss = (
                EmpiricalDistribution(
                    samples=reconciled_samples_at_all_levels, event_dim=1
                )
                .loss(x=target_at_all_levels)
                .expand_dims(axis=-1)
            )

        return loss


class COPDeepARPredictionNetwork(COPNetwork):
    @validated()
    def __init__(
        self,
        return_forecasts_at_all_levels: bool = False,
        num_parallel_samples: int = 100,
        **kwargs,
    ) -> None:
        super().__init__(prediction=True, **kwargs)
        self.return_forecasts_at_all_levels = return_forecasts_at_all_levels
        self.num_parallel_samples = num_parallel_samples

    def _decode_one_window(
        self,
        F,
        model: DeepARPredictionNetwork,
        window_size: int,
        offset: int,
        static_feat: Tensor,
        past_target: Tensor,
        time_feat: Tensor,
        scale: Tensor,
        begin_states: List,
    ) -> Tuple[Tensor, Tensor]:
        """
        Computes RNN outputs by unrolling the LSTM starting with a initial
        input and state.

        Parameters
        ----------
        static_feat : Tensor
            static features. Shape: (batch_size, num_static_features).
        past_target : Tensor
            target history. Shape: (batch_size, history_length).
        time_feat : Tensor
            time features. Shape: (batch_size, prediction_length,
            num_time_features).
            Note: They still need to be for all `prediction_length` time steps.
            This function will slice the features it needs.
        scale : Tensor
            tensor containing the scale of each element in the batch.
            Shape: (batch_size, 1, 1).
        begin_states : List
            list of initial states for the LSTM layers. The shape of each
            tensor of the list should be (batch_size, num_cells)
        Returns
        --------
        Tensor
            A tensor containing sampled paths.
            Shape: (batch_size, num_sample_paths, window_size).
        """

        rnn_outputs_ls = []

        # for each future time-units we draw new samples for this time-unit and
        # update the state
        for k in range(offset, offset + window_size):
            # (batch_size * num_samples, 1, *target_shape, num_lags)
            lags = model.get_lagged_subsequences(
                F=F,
                sequence=past_target,
                sequence_length=model.history_length + k,
                indices=model.shifted_lags,
                subsequences_length=1,
            )

            # (batch_size * num_samples, 1, *target_shape, num_lags)
            lags_scaled = F.broadcast_div(lags, scale.expand_dims(axis=-1))

            # from (batch_size * num_samples, 1, *target_shape, num_lags)
            # to (batch_size * num_samples, 1, prod(target_shape) * num_lags)
            input_lags = F.reshape(
                data=lags_scaled,
                shape=(-1, 1, prod(model.target_shape) * len(model.lags_seq)),
            )

            # (batch_size * num_samples, 1, prod(target_shape) * num_lags +
            # num_time_features + num_static_features)
            decoder_input = F.concat(
                input_lags,
                time_feat.slice_axis(axis=1, begin=k, end=k + 1),
                # observed_values.expand_dims(axis=1),
                static_feat,
                dim=-1,
            )

            # output shape: (batch_size * num_samples, 1, num_cells)
            # state shape: (batch_size * num_samples, num_cells)
            rnn_outputs, begin_states = model.rnn.unroll(
                inputs=decoder_input,
                length=1,
                begin_state=begin_states,
                layout="NTC",
                merge_outputs=True,
            )
            rnn_outputs_ls.append(rnn_outputs)

            distr_args = model.proj_distr_args(rnn_outputs)

            # compute likelihood of target given the predicted parameters
            distr = model.distr_output.distribution(distr_args, scale=scale)

            # (batch_size * num_samples, 1, *target_shape)

            new_samples = distr.sample(dtype=self.dtype)

            # (batch_size * num_samples, seq_len, *target_shape)
            past_target = F.concat(past_target, new_samples, dim=1)

        # (batch_size * num_samples, prediction_length, *target_shape)
        rnn_outputs = F.concat(*rnn_outputs_ls, dim=1)

        return rnn_outputs, begin_states

    def sampling_decoder(
        self,
        F,
        state_ls,
        scale_ls,
        static_feat_ls,
        past_target_ls,
        future_time_feat_agg_ls,
    ):
        num_windows = (
            self.prediction_length // self.temporal_hierarchy.num_leaves
        )
        num_nodes_per_level = self.temporal_hierarchy.num_nodes_per_level

        reconciled_samples_at_all_levels_ls = []
        for j in range(num_windows):
            embeddings_at_all_levels_ls = []
            for i, agg_multiple in enumerate(
                self.temporal_hierarchy.agg_multiples
            ):
                rnn_outputs, states = self._decode_one_window(
                    F=F,
                    model=self.models[i],
                    window_size=num_nodes_per_level[i],
                    offset=j * num_nodes_per_level[i],
                    past_target=past_target_ls[i],
                    time_feat=future_time_feat_agg_ls[i],
                    static_feat=static_feat_ls[i],
                    scale=scale_ls[i],
                    begin_states=state_ls[i],
                )

                state_ls[i] = states
                embeddings_at_all_levels_ls.append(
                    rnn_outputs.reshape(
                        (rnn_outputs.shape[0], -1, rnn_outputs.shape[-1])
                    )
                )

            # Last dimension contains embeddings at all time-levels and message passing/aggregation can be done on it.
            # Shape: (bs, total_num_time_steps_of_hierarchy, embedding_dim)
            embeddings_at_all_levels = F.concat(
                *embeddings_at_all_levels_ls, dim=-2
            )

            if self.use_gnn:
                embeddings_at_all_levels = self.gnn(embeddings_at_all_levels)

            distr_at_all_levels = self._embeddings_to_distr(
                F,
                embeddings_at_all_levels,
                scale_ls,
            )

            samples_at_all_levels = self._distr_to_samples(
                distr_at_all_levels,
                num_samples=1,
            )

            if self.do_reconciliation:
                reconciled_samples_at_all_levels = reconcile_samples(
                    reconciliation_mat=self.M,
                    samples=samples_at_all_levels,
                    non_negative=self.non_negative,
                )
            else:
                reconciled_samples_at_all_levels = samples_at_all_levels

            rec_err = coherency_error(
                S=self.temporal_hierarchy.agg_mat,
                samples=reconciled_samples_at_all_levels.asnumpy(),
            )
            print(f"Reconciliation error: {rec_err}")

            cumsum_nodes_per_level = np.cumsum([0] + num_nodes_per_level)
            for i in range(len(self.temporal_hierarchy.agg_multiples)):
                # (batch_size * num_samples, seq_len, *target_shape)
                reconciled_samples = (
                    reconciled_samples_at_all_levels.slice_axis(
                        axis=-1,
                        begin=cumsum_nodes_per_level[i],
                        end=cumsum_nodes_per_level[i + 1],
                    )
                )
                past_target_ls[i] = F.concat(
                    past_target_ls[i], reconciled_samples, dim=1
                )

            reconciled_samples_at_all_levels_ls.append(
                reconciled_samples_at_all_levels.reshape(
                    shape=(
                        -1,
                        self.num_parallel_samples,
                        reconciled_samples_at_all_levels.shape[-1],
                    )
                ).expand_dims(axis=-2)
            )

        reconciled_samples_at_all_levels = F.concat(
            *reconciled_samples_at_all_levels_ls, dim=-2
        )
        print(reconciled_samples_at_all_levels.shape)
        return reconciled_samples_at_all_levels

    # noinspection PyMethodOverriding,PyPep8Naming
    def hybrid_forward(
        self,
        F,
        feat_static_cat: Tensor,  # (batch_size, num_features)
        feat_static_real: Tensor,  # (batch_size, num_features)
        past_time_feat: Tensor,  # (batch_size, history_length, num_features)
        past_target: Tensor,  # (batch_size, history_length, *target_shape)
        past_observed_values: Tensor,  # (batch_size, history_length, *target_shape)
        future_time_feat: Tensor,  # (batch_size, prediction_length, num_features)
        past_is_pad: Tensor,
        agg_features_dict: Dict,
    ) -> Tensor:
        """
        Predicts samples, all tensors should have NTC layout.
        Parameters
        ----------
        F
        feat_static_cat : (batch_size, num_features)
        feat_static_real : (batch_size, num_features)
        past_time_feat : (batch_size, history_length, num_features)
        past_target : (batch_size, history_length, *target_shape)
        past_observed_values : (batch_size, history_length, *target_shape)
        future_time_feat : (batch_size, prediction_length, num_features)
        agg_features_dict: Dictionary of features for aggregated levels

        Returns
        -------
        Tensor
            Predicted samples
        """

        (
            state_ls,
            scale_ls,
            static_feat_ls,
            past_target_ls,
            future_time_feat_agg_ls,
        ) = ([], [], [], [], [])

        for i, agg_multiple in enumerate(
            self.temporal_hierarchy.agg_multiples
        ):
            if agg_multiple != 1:
                past_time_feat_agg = agg_features_dict[f"level_{i}"][
                    "past_time_feat_agg"
                ]
                future_time_feat_agg = agg_features_dict[f"level_{i}"][
                    "future_time_feat_agg"
                ]
            else:
                past_time_feat_agg = past_time_feat
                future_time_feat_agg = future_time_feat

            target_related_feat_agg = (
                self.get_target_related_feat_at_agg_level(
                    agg_level=i,
                    past_target=past_target,
                    past_is_pad=past_is_pad,
                    past_observed_values=past_observed_values,
                )
            )

            # unroll the decoder in "prediction mode", i.e. with past data only
            _, states, scale, static_feat, imputed_sequence = self.models[
                i
            ].unroll_encoder(
                F=F,
                feat_static_cat=feat_static_cat,
                feat_static_real=feat_static_real,
                past_time_feat=past_time_feat_agg,
                future_observed_values=None,
                future_time_feat=None,
                future_target=None,
                **target_related_feat_agg,
            )

            # blows-up the dimension of each tensor to batch_size *
            # self.num_parallel_samples for increasing parallelism
            repeated_past_target = imputed_sequence.repeat(
                repeats=self.num_parallel_samples, axis=0
            )
            repeated_states = [
                s.repeat(repeats=self.num_parallel_samples, axis=0)
                for s in states
            ]
            repeated_time_feat = future_time_feat_agg.repeat(
                repeats=self.num_parallel_samples, axis=0
            )
            repeated_static_feat = static_feat.repeat(
                repeats=self.num_parallel_samples, axis=0
            ).expand_dims(axis=1)
            repeated_scale = scale.repeat(
                repeats=self.num_parallel_samples, axis=0
            )

            state_ls.append(repeated_states)
            scale_ls.append(repeated_scale)
            static_feat_ls.append(repeated_static_feat)
            past_target_ls.append(repeated_past_target)
            future_time_feat_agg_ls.append(repeated_time_feat)

        reconciled_samples_at_all_levels = self.sampling_decoder(
            F,
            state_ls=state_ls,
            scale_ls=scale_ls,
            static_feat_ls=static_feat_ls,
            past_target_ls=past_target_ls,
            future_time_feat_agg_ls=future_time_feat_agg_ls,
        )

        if self.return_forecasts_at_all_levels:
            return reconciled_samples_at_all_levels
        else:
            reconciled_samples_at_bottom_level = (
                reconciled_samples_at_all_levels.slice_axis(
                    axis=-1,
                    begin=-self.temporal_hierarchy.num_leaves,
                    end=None,
                )
            )

            reconciled_samples_at_bottom_level = (
                reconciled_samples_at_bottom_level.reshape(
                    (
                        reconciled_samples_at_bottom_level.shape[0],
                        reconciled_samples_at_bottom_level.shape[1],
                        -1,
                    )
                )
            )

            return reconciled_samples_at_bottom_level



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import json
from dataclasses import dataclass
from pathlib import Path
from typing import cast, List, Optional
from gluonts_customized.mx import DeepAREstimator
from gluonts_customized.model.estimator import DummyEstimator, Estimator
from gluonts_customized.mx import NBEATSEstimator
from gluonts_customized.model.naive_2 import Naive2Predictor
from gluonts_customized.model.npts import NPTSEstimator
from gluonts_customized.model.predictor import Predictor
from gluonts_customized.model.prophet import ProphetPredictor
from gluonts_customized.model.r_forecast import RForecastPredictor
from gluonts_customized.model.seasonal_naive import SeasonalNaivePredictor
from gluonts_customized.mx import MQCNNEstimator, MQRNNEstimator
from gluonts_customized.mx import SimpleFeedForwardEstimator
from gluonts_customized.mx.model.tft import TemporalFusionTransformerEstimator
from gluonts_customized.mx.trainer.callback import Callback
from gluonts_customized.time_feature import Constant
from mxnet.gluon import nn
from tsbench.config.dataset import DatasetConfig
from tsbench.config.dataset.datasets import WindFarmsDatasetConfig
from ._base import ModelConfig, TrainConfig
from ._factory import register_model


@register_model
@dataclass(frozen=True)
class SeasonalNaiveModelConfig(ModelConfig):
    """
    The seasonal naive estimator.
    """

    @classmethod
    def name(cls) -> str:
        return "seasonal_naive"

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return DummyEstimator(
            SeasonalNaivePredictor,
            freq=freq,
            prediction_length=prediction_length,
        )


@register_model
@dataclass(frozen=True)
class ThetaModelConfig(ModelConfig):
    """
    The Theta R estimator config.
    """

    @classmethod
    def name(cls) -> str:
        return "theta"

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return DummyEstimator(
            RForecastPredictor,
            freq=freq,
            prediction_length=prediction_length,
            method_name="thetaf",
        )


@dataclass(frozen=True)
class Naive2ModelConfig(ModelConfig):
    """
    The Naïve 2 model as used in the M4 competition.
    """

    @classmethod
    def name(cls) -> str:
        return "naive2"

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return DummyEstimator(
            Naive2Predictor, freq=freq, prediction_length=prediction_length
        )


@register_model
@dataclass(frozen=True)
class NptsModelConfig(ModelConfig):
    """
    The NPTS estimator config.
    """

    @classmethod
    def name(cls) -> str:
        return "npts"

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return NPTSEstimator(freq=freq, prediction_length=prediction_length)


@register_model
@dataclass(frozen=True)
class DeepARModelConfig(ModelConfig, TrainConfig):
    """
    The DeepAR estimator config.
    """

    num_layers: int = 2
    num_cells: int = 40

    @classmethod
    def name(cls) -> str:
        return "deepar"

    def create_predictor(
        self, estimator: Estimator, network: nn.HybridBlock
    ) -> Predictor:
        deepar_estimator = cast(DeepAREstimator, estimator)
        transform = deepar_estimator.create_transformation()
        return deepar_estimator.create_predictor(transform, network)

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return DeepAREstimator(
            freq=freq,
            prediction_length=prediction_length,
            num_layers=self.num_layers,
            num_cells=self.num_cells,
            time_features=None if time_features else [],
            trainer=self._create_trainer(
                training_time,
                validation_milestones,
                callbacks,  # type: ignore
            ),
            context_length=self.context_length_multiple * prediction_length,
        )


@register_model
@dataclass(frozen=True)
class MQCnnModelConfig(ModelConfig, TrainConfig):
    """
    The MQCNN estimator config.
    """

    context_length_multiple: int = 4
    num_filters: int = 30
    kernel_size_first: int = 7
    kernel_size_hidden: int = 3
    kernel_size_last: int = 3

    @classmethod
    def name(cls) -> str:
        return "mqcnn"

    def create_predictor(
        self, estimator: Estimator, network: nn.HybridBlock
    ) -> Predictor:
        mqcnn_estimator = cast(MQCNNEstimator, estimator)
        transform = mqcnn_estimator.create_transformation()
        return mqcnn_estimator.create_predictor(transform, cast)  # type: ignore

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return MQCNNEstimator(
            freq=freq,
            prediction_length=prediction_length,
            channels_seq=[self.num_filters] * 3,
            kernel_size_seq=[
                self.kernel_size_first,
                self.kernel_size_hidden,
                self.kernel_size_last,
            ],
            add_time_feature=time_features,
            trainer=self._create_trainer(
                training_time,
                validation_milestones,
                callbacks,  # type: ignore
            ),
            context_length=self.context_length_multiple * prediction_length,
        )


@register_model
@dataclass(frozen=True)
class MQRnnModelConfig(ModelConfig, TrainConfig):
    """
    The MQRNN estimator config.
    """

    context_length_multiple: int = 4

    @classmethod
    def name(cls) -> str:
        return "mqrnn"

    def create_predictor(
        self, estimator: Estimator, network: nn.HybridBlock
    ) -> Predictor:
        mqrnn_estimator = cast(MQRNNEstimator, estimator)
        transform = mqrnn_estimator.create_transformation()
        return mqrnn_estimator.create_predictor(transform, network)  # type: ignore

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return MQRNNEstimator(
            freq=freq,
            prediction_length=prediction_length,
            trainer=self._create_trainer(
                training_time,
                validation_milestones,
                callbacks,  # type: ignore
            ),
            context_length=self.context_length_multiple * prediction_length,
        )


@register_model
@dataclass(frozen=True)
class SimpleFeedforwardModelConfig(ModelConfig, TrainConfig):
    """
    The simple feedforward estimator config.
    """

    hidden_dim: int = 40
    num_layers: int = 2

    @classmethod
    def name(cls) -> str:
        return "simple_feedforward"

    def create_predictor(
        self, estimator: Estimator, network: nn.HybridBlock
    ) -> Predictor:
        ff_estimator = cast(SimpleFeedForwardEstimator, estimator)
        transform = ff_estimator.create_transformation()
        return ff_estimator.create_predictor(transform, network)

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return SimpleFeedForwardEstimator(
            freq=freq,
            prediction_length=prediction_length,
            num_hidden_dimensions=[self.hidden_dim] * self.num_layers,
            trainer=self._create_trainer(
                training_time,
                validation_milestones,
                callbacks,  # type: ignore
            ),
            context_length=self.context_length_multiple * prediction_length,
        )


@register_model
@dataclass(frozen=True)
class TemporalFusionTransformerModelConfig(ModelConfig, TrainConfig):
    """
    The temporal fusion transformer estimator config.
    """

    hidden_dim: int = 32
    num_heads: int = 4

    @classmethod
    def name(cls) -> str:
        return "tft"

    def create_predictor(
        self, estimator: Estimator, network: nn.HybridBlock
    ) -> Predictor:
        tft_estimator = cast(TemporalFusionTransformerEstimator, estimator)
        transform = tft_estimator.create_transformation()
        return tft_estimator.create_predictor(transform, network)

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return TemporalFusionTransformerEstimator(
            freq=freq,
            prediction_length=prediction_length,
            hidden_dim=self.hidden_dim,
            num_heads=self.num_heads,
            num_outputs=9,
            time_features=[Constant()] if not time_features else [],
            trainer=self._create_trainer(
                training_time,
                validation_milestones,
                callbacks,  # type: ignore
            ),
        )


@register_model
@dataclass(frozen=True)
class NBeatsModelConfig(ModelConfig, TrainConfig):
    """
    The NBeats estimator config.
    """

    context_length_multiple: int = 2
    num_stacks: int = 30
    num_blocks: int = 1

    @classmethod
    def name(cls) -> str:
        return "nbeats"

    @property
    def prediction_samples(self) -> int:
        return 1

    def create_predictor(
        self, estimator: Estimator, network: nn.HybridBlock
    ) -> Predictor:
        nb_estimator = cast(NBEATSEstimator, estimator)
        transform = nb_estimator.create_transformation()
        return nb_estimator.create_predictor(transform, network)

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return NBEATSEstimator(
            freq=freq,
            prediction_length=prediction_length,
            num_stacks=self.num_stacks,
            num_blocks=[self.num_blocks],
            trainer=self._create_trainer(
                training_time,
                validation_milestones,
                callbacks,  # type: ignore
            ),
            context_length=self.context_length_multiple * prediction_length,
        )


@register_model
@dataclass(frozen=True)
class ProphetModelConfig(ModelConfig):
    """
    The Prophet estimator config.
    """

    @classmethod
    def name(cls) -> str:
        return "prophet"

    def save_predictor(self, predictor: Predictor, path: Path) -> None:
        file = path / "metadata.pickle"
        with file.open("w") as f:
            json.dump(
                {
                    "freq": predictor.freq,
                    "prediction_length": predictor.prediction_length,
                },
                f,
            )

    def load_predictor(self, path: Path) -> Predictor:
        file = path / "metadata.pickle"
        with file.open("r") as f:
            meta = json.load(f)
        return ProphetPredictor(
            freq=meta["freq"], prediction_length=meta["prediction_length"]
        )

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return DummyEstimator(
            ProphetPredictor, freq=freq, prediction_length=prediction_length
        )


@register_model
@dataclass(frozen=True)
class STLARModelConfig(ModelConfig):
    """
    The STL-AR estimator config.
    """

    @classmethod
    def name(cls) -> str:
        return "stlar"

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return DummyEstimator(
            RForecastPredictor,
            freq=freq,
            prediction_length=prediction_length,
            method_name="stlar",
        )


@register_model
@dataclass(frozen=True)
class ARIMAModelConfig(ModelConfig):
    """
    The ARIMA estimator config.
    """

    @classmethod
    def name(cls) -> str:
        return "arima"

    @property
    def prefers_parallel_predictions(self) -> bool:
        return True

    def max_time_series_length(self, config: DatasetConfig) -> Optional[int]:
        if isinstance(config, WindFarmsDatasetConfig):
            return 2880
        return 100000

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return DummyEstimator(
            RForecastPredictor,
            freq=freq,
            prediction_length=prediction_length,
            method_name="arima",
        )


@register_model
@dataclass(frozen=True)
class ETSModelConfig(ModelConfig):
    """
    The ETS estimator config.
    """

    @classmethod
    def name(cls) -> str:
        return "ets"

    def create_estimator(
        self,
        freq: str,
        prediction_length: int,
        time_features: bool,
        training_time: float,
        validation_milestones: List[float],
        callbacks: List[Callback],
    ) -> Estimator:
        return DummyEstimator(
            RForecastPredictor,
            freq=freq,
            prediction_length=prediction_length,
            method_name="ets",
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import os
import warnings
from pathlib import Path
from typing import NamedTuple, Optional

import numpy as np
import pandas as pd

from gluonts_customized.dataset import DatasetWriter
from gluonts_customized.dataset.common import MetaData, TrainDatasets
from gluonts_customized.dataset.repository._util import metadata
from gluonts_customized.gluonts_tqdm import tqdm


def check_dataset(dataset_path: Path, length: int, sheet_name):
    # check that things are correct
    from gluonts_customized.dataset.common import load_datasets

    ds = load_datasets(
        metadata=dataset_path,
        train=dataset_path / "train",
        test=dataset_path / "test",
    )

    assert ds.test is not None
    assert len(list(ds.train)) == length
    assert len(list(ds.test)) == length

    assert ds.metadata.prediction_length is not None

    for ts_train, ts_test in tqdm(
        zip(ds.train, ds.test), total=length, desc="checking consistency"
    ):
        train_target = ts_train["target"]
        test_target = ts_test["target"]
        assert (
            len(train_target)
            == len(test_target) - ds.metadata.prediction_length
        )
        assert np.all(train_target == test_target[: len(train_target)])

        assert ts_train["start"] == ts_test["start"]
        start = ts_train["start"]

        if sheet_name in ["M3Quart", "Other"]:
            assert (start.month, start.day) in [
                (3, 31),
                (6, 30),
                (9, 30),
                (12, 31),
            ], f"Invalid time stamp {start.month}-{start.day}"
        elif sheet_name == "M3Year":
            assert (start.month, start.day) == (
                12,
                31,
            ), f"Invalid time stamp {start.month}-{start.day}"


class M3Setting(NamedTuple):
    sheet_name: str
    prediction_length: int
    freq: str


def generate_m3_dataset(
    dataset_path: Path,
    m3_freq: str,
    dataset_writer: DatasetWriter,
    prediction_length: Optional[int] = None,
):
    from gluonts_customized.dataset.repository import default_dataset_path

    m3_xls_path = default_dataset_path / "M3C.xls"
    if not os.path.exists(m3_xls_path):
        raise RuntimeError(
            "The m3 data is available at "
            "https://forecasters.org/resources/"
            "time-series-data/m3-competition/ "
            "Please download the file and copy the files to this location: "
            f"{m3_xls_path}"
        )

    subsets = {
        "yearly": M3Setting("M3Year", 6, "Y"),
        "quarterly": M3Setting("M3Quart", 8, "Q"),
        "monthly": M3Setting("M3Month", 18, "M"),
        "other": M3Setting("M3Other", 8, "Q"),
    }
    assert (
        m3_freq.lower() in subsets
    ), f"invalid m3_freq='{m3_freq}'. Allowed values: {subsets.keys()}"

    if m3_freq.lower() == "other":
        warnings.warn(
            "Be aware: The M3-other dataset does not have a known frequency."
            " Since gluonts needs a known frequency, we will generate the"
            " dataset with an artificial `quarterly` frequency."
        )

    subset = subsets[m3_freq.lower()]
    df = pd.read_excel(m3_xls_path, sheet_name=subset.sheet_name)

    def truncate_trailing_nan(v: np.ndarray):
        last_finite_index = np.where(np.isfinite(v))[0][-1]
        return v[: last_finite_index + 1]

    train_data = []
    test_data = []

    def normalize_category(c: str):
        return c.strip()

    df["Category"] = df["Category"].apply(normalize_category)
    categories = list(df["Category"].unique())

    cat_map = {c: i for i, c in enumerate(categories)}

    for i, (_, row) in enumerate(df.iterrows()):
        vals = row.values
        series, n, nf, category, starting_year, starting_offset = vals[:6]
        target = np.asarray(vals[6:], dtype=np.float64)
        target = truncate_trailing_nan(target)
        assert len(target) == n
        assert nf == subset.prediction_length
        mock_start = "1750"

        if starting_year == 0:
            assert starting_offset == 0
            starting_year = mock_start

        # fix bugs in M3C xls
        if series == "N1071":
            # bug in the m3 dataset
            starting_offset = 1
        if series == "N 184":
            starting_offset = 1

        offset = max(starting_offset - 1, 0)

        if subset.freq == "Q":
            assert 0 <= offset < 4
            time_stamp = f"{starting_year}-{3 * (offset + 1):02}-15"
        elif subset.freq == "Y":
            assert offset == 0
            time_stamp = f"{starting_year}-12-15"
        elif subset.freq == "M":
            assert 0 <= offset < 12
            time_stamp = f"{starting_year}-{offset + 1:02}-15"

        start = str(pd.Period(time_stamp, freq=subset.freq))
        cat = [i, cat_map[category]]

        train_data.append(
            {
                "target": target[: -subset.prediction_length],
                "start": start,
                "feat_static_cat": cat,
                "item_id": series,
            }
        )

        test_data.append(
            {
                "target": target,
                "start": start,
                "feat_static_cat": cat,
                "item_id": series,
            }
        )

    meta = MetaData(
        **metadata(
            cardinality=[len(train_data), len(categories)],
            freq=subset.freq,
            prediction_length=prediction_length or subset.prediction_length,
        )
    )

    dataset = TrainDatasets(metadata=meta, train=train_data, test=test_data)
    dataset.save(
        path_str=str(dataset_path), writer=dataset_writer, overwrite=True
    )

    check_dataset(dataset_path, len(df), subset.sheet_name)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Tuple, Union

import numpy as np
import pytest
import pandas as pd

from gluonts_customized.dataset.split import split, TestData
from gluonts_customized.itertools import prod
from gluonts_customized.model.seasonal_naive import SeasonalNaivePredictor
from gluonts_customized.model import (
    SampleForecast,
    evaluate_forecasts,
    evaluate_model,
)
from gluonts_customized.ev.metrics import (
    MetricDefinition,
    MSE,
    RMSE,
    NRMSE,
    MAPE,
    SMAPE,
    MASE,
    MSIS,
    ND,
    WeightedSumQuantileLoss,
    MeanWeightedSumQuantileLoss,
    MeanScaledQuantileLoss,
    AverageMeanScaledQuantileLoss,
)


_test_univariate_dataset = [
    {
        "item_id": k,
        "start": pd.Period("2022-06-12", freq="D"),
        "target": np.ones((100)),
    }
    for k in range(5)
]

_test_multivariate_dataset = [
    {
        "item_id": k,
        "start": pd.Period("2022-06-12", freq="D"),
        "target": np.ones((7, 100)),
    }
    for k in range(5)
]

_test_metrics = [
    MSE(),
    RMSE(),
    NRMSE(),
    MAPE(),
    SMAPE(),
    MASE(),
    MSIS(),
    ND(),
    WeightedSumQuantileLoss(0.5),
    MeanWeightedSumQuantileLoss(quantile_levels=[0.1, 0.5, 0.9]),
    MeanScaledQuantileLoss(0.5),
    AverageMeanScaledQuantileLoss(quantile_levels=[0.1, 0.5, 0.9]),
]


def infer_metrics_df_shape(
    metrics: List[MetricDefinition],
    test_data: TestData,
    axis: Union[int, Tuple[int]],
) -> Tuple[int]:
    """
    Infer expected shape of metrics data frame.
    """
    labels = list(test_data.label)
    test_data_shape = (len(labels), *labels[0]["target"].shape)

    if axis is None:
        axis = tuple(range(len(test_data_shape)))
    if isinstance(axis, int):
        axis = (axis,)

    return (
        prod(
            test_data_shape[d]
            for d in range(len(test_data_shape))
            if d not in axis
        ),
        len(metrics),
    )


@pytest.mark.parametrize(
    "metrics, test_data, axis",
    [
        (
            _test_metrics,
            split(_test_univariate_dataset, offset=-12)[1].generate_instances(
                prediction_length=3, windows=4
            ),
            None,
        ),
        (
            _test_metrics,
            split(_test_univariate_dataset, offset=-12)[1].generate_instances(
                prediction_length=3, windows=4
            ),
            (0, 1),
        ),
        (
            _test_metrics,
            split(_test_univariate_dataset, offset=-12)[1].generate_instances(
                prediction_length=3, windows=4
            ),
            0,
        ),
        (
            _test_metrics,
            split(_test_univariate_dataset, offset=-12)[1].generate_instances(
                prediction_length=3, windows=4
            ),
            1,
        ),
        (
            _test_metrics,
            split(_test_multivariate_dataset, offset=-12)[
                1
            ].generate_instances(prediction_length=3, windows=4),
            None,
        ),
        (
            _test_metrics,
            split(_test_multivariate_dataset, offset=-12)[
                1
            ].generate_instances(prediction_length=3, windows=4),
            (0, 1, 2),
        ),
        (
            _test_metrics,
            split(_test_multivariate_dataset, offset=-12)[
                1
            ].generate_instances(prediction_length=3, windows=4),
            (0, 2),
        ),
        (
            _test_metrics,
            split(_test_multivariate_dataset, offset=-12)[
                1
            ].generate_instances(prediction_length=3, windows=4),
            (0, 1),
        ),
        (
            _test_metrics,
            split(_test_multivariate_dataset, offset=-12)[
                1
            ].generate_instances(prediction_length=3, windows=4),
            (1, 2),
        ),
    ],
)
def test_evaluation_shape(
    metrics: List[MetricDefinition],
    test_data: TestData,
    axis: Union[int, Tuple[int]],
):
    num_samples = 31

    sample_forecasts = [
        SampleForecast(
            samples=np.zeros((num_samples, *label["target"].T.shape)),
            start_date=label["start"],
            item_id=label["item_id"],
        )
        for label in test_data.label
    ]

    metrics_df = evaluate_forecasts(
        sample_forecasts,
        test_data=test_data,
        metrics=metrics,
        axis=axis,
    )

    assert metrics_df.shape == infer_metrics_df_shape(metrics, test_data, axis)


def test_evaluate_model_vs_forecasts():
    dataset = [
        {
            "item_id": k,
            "start": pd.Period("2022-06-12", freq="D"),
            "target": np.random.normal(size=50),
        }
        for k in range(2)
    ]

    test_data = split(dataset, offset=-12)[1].generate_instances(
        prediction_length=3, windows=4
    )

    model = SeasonalNaivePredictor(prediction_length=3, season_length=1)

    forecasts = list(model.predict(test_data.input))

    for axis in [None, 0, 1, (0, 1)]:
        df1 = evaluate_forecasts(
            forecasts=forecasts,
            test_data=test_data,
            metrics=_test_metrics,
            axis=axis,
        )
        df2 = evaluate_model(
            model=model, test_data=test_data, metrics=_test_metrics, axis=axis
        )
        assert (df1 == df2).all().all()


def test_data_nan():
    target = np.random.normal(size=50)
    target[0] = np.nan
    target[-1] = np.nan
    dataset = [
        {
            "item_id": k,
            "start": pd.Period("2022-06-12", freq="D"),
            "target": target,
        }
        for k in range(2)
    ]

    test_data = split(dataset, offset=-12)[1].generate_instances(
        prediction_length=3, windows=4
    )

    model = SeasonalNaivePredictor(prediction_length=3, season_length=1)

    forecasts = list(model.predict(test_data.input))

    for axis in [None, 0, 1, (0, 1)]:
        df1 = evaluate_forecasts(
            forecasts=forecasts,
            test_data=test_data,
            metrics=_test_metrics,
            axis=axis,
        )
        assert not np.any(df1.isna())



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Tuple

import mxnet as mx

from gluonts_customized.core.component import validated
from gluonts_customized.mx import Tensor
from gluonts_customized.mx.block.scaler import MeanScaler, NOPScaler
from gluonts_customized.mx.distribution import DistributionOutput
from gluonts_customized.mx.util import weighted_average


class SimpleFeedForwardNetworkBase(mx.gluon.HybridBlock):
    """
    Abstract base class to implement feed-forward networks for probabilistic
    time series prediction.

    This class does not implement hybrid_forward: this is delegated
    to the two subclasses SimpleFeedForwardTrainingNetwork and
    SimpleFeedForwardPredictionNetwork, that define respectively how to
    compute the loss and how to generate predictions.

    Parameters
    ----------
    num_hidden_dimensions
        Number of hidden nodes in each layer.
    prediction_length
        Number of time units to predict.
    context_length
        Number of time units that condition the predictions.
    batch_normalization
        Whether to use batch normalization.
    mean_scaling
        Scale the network input by the data mean and the network output by
        its inverse.
    distr_output
        Distribution to fit.
    kwargs
    """

    # Needs the validated decorator so that arguments types are checked and
    # the block can be serialized.
    @validated()
    def __init__(
        self,
        num_hidden_dimensions: List[int],
        prediction_length: int,
        context_length: int,
        batch_normalization: bool,
        mean_scaling: bool,
        distr_output: DistributionOutput,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.num_hidden_dimensions = num_hidden_dimensions
        self.prediction_length = prediction_length
        self.context_length = context_length
        self.batch_normalization = batch_normalization
        self.mean_scaling = mean_scaling
        self.distr_output = distr_output

        with self.name_scope():
            self.distr_args_proj = self.distr_output.get_args_proj()
            self.mlp = mx.gluon.nn.HybridSequential()
            dims = self.num_hidden_dimensions
            for layer_no, units in enumerate(dims[:-1]):
                self.mlp.add(mx.gluon.nn.Dense(units=units, activation="relu"))
                if self.batch_normalization:
                    self.mlp.add(mx.gluon.nn.BatchNorm())
            self.mlp.add(mx.gluon.nn.Dense(units=prediction_length * dims[-1]))
            self.mlp.add(
                mx.gluon.nn.HybridLambda(
                    lambda F, o: F.reshape(
                        o, (-1, prediction_length, dims[-1])
                    )
                )
            )
            self.scaler = MeanScaler() if mean_scaling else NOPScaler()

    def get_distr_args(
        self, F, past_target: Tensor
    ) -> Tuple[Tensor, Tensor, Tensor]:
        """
        Given past target values, applies the feed-forward network and maps the
        output to the parameter of probability distribution for future
        observations.

        Parameters
        ----------
        F
        past_target
            Tensor containing past target observations.
            Shape: (batch_size, context_length, target_dim).

        Returns
        -------
        Tensor
            The parameters of distribution.
        Tensor
            An array containing the location (shift) of the distribution.
        Tensor
            An array containing the scale of the distribution.
        """
        scaled_target, target_scale = self.scaler(
            past_target,
            F.ones_like(past_target),
        )
        mlp_outputs = self.mlp(scaled_target)
        distr_args = self.distr_args_proj(mlp_outputs)
        scale = target_scale.expand_dims(axis=1)
        loc = F.zeros_like(scale)
        return distr_args, loc, scale


class SimpleFeedForwardTrainingNetwork(SimpleFeedForwardNetworkBase):
    def hybrid_forward(
        self,
        F,
        past_target: Tensor,
        future_target: Tensor,
        future_observed_values: Tensor,
    ) -> Tensor:
        """
        Computes a probability distribution for future data given the past, and
        returns the loss associated with the actual future observations.

        Parameters
        ----------
        F
        past_target
            Tensor with past observations.
            Shape: (batch_size, context_length, target_dim).
        future_target
            Tensor with future observations.
            Shape: (batch_size, prediction_length, target_dim).
        future_observed_values
            Tensor indicating which values in the target are observed, and
            which ones are imputed instead.

        Returns
        -------
        Tensor
            Loss tensor. Shape: (batch_size, ).
        """
        distr_args, loc, scale = self.get_distr_args(F, past_target)
        distr = self.distr_output.distribution(
            distr_args, loc=loc, scale=scale
        )

        # (batch_size, prediction_length, target_dim)
        loss = distr.loss(future_target)

        weighted_loss = weighted_average(
            F=F, x=loss, weights=future_observed_values, axis=1
        )

        # (batch_size, )
        return weighted_loss


class SimpleFeedForwardSamplingNetwork(SimpleFeedForwardNetworkBase):
    @validated()
    def __init__(
        self, num_parallel_samples: int = 100, *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.num_parallel_samples = num_parallel_samples

    def hybrid_forward(self, F, past_target: Tensor) -> Tensor:
        """
        Computes a probability distribution for future data given the past, and
        draws samples from it.

        Parameters
        ----------
        F
        past_target
            Tensor with past observations.
            Shape: (batch_size, context_length, target_dim).

        Returns
        -------
        Tensor
            Prediction sample. Shape: (batch_size, samples, prediction_length).
        """

        distr_args, loc, scale = self.get_distr_args(F, past_target)
        distr = self.distr_output.distribution(
            distr_args, loc=loc, scale=scale
        )

        # (num_samples, batch_size, prediction_length)
        samples = distr.sample(self.num_parallel_samples)

        # (batch_size, num_samples, prediction_length)
        return samples.swapaxes(0, 1)


class SimpleFeedForwardDistributionNetwork(SimpleFeedForwardNetworkBase):
    @validated()
    def __init__(
        self, num_parallel_samples: int = 100, *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.num_parallel_samples = num_parallel_samples

    def hybrid_forward(self, F, past_target: Tensor) -> Tensor:
        """
        Computes the parameters of distribution for future data given the past,
        and draws samples from it.

        Parameters
        ----------
        F
        past_target
            Tensor with past observations.
            Shape: (batch_size, context_length, target_dim).

        Returns
        -------
        Tensor
            The parameters of distribution.
        Tensor
            An array containing the location (shift) of the distribution.
        Tensor
            An array containing the scale of the distribution.
        """
        distr_args, loc, scale = self.get_distr_args(F, past_target)
        return distr_args, loc, scale



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import math

import mxnet as mx
import mxnet.gluon.nn as nn
import numpy as np
import pandas as pd
import pytest

from gluonts_customized.dataset.common import ListDataset
from gluonts_customized.mx import SimpleFeedForwardEstimator
from gluonts_customized.mx.trainer import Trainer
from gluonts_customized.mx.trainer.model_iteration_averaging import (
    NTA,
    Alpha_Suffix,
)


def initialize_model() -> nn.HybridBlock:
    # dummy training data
    N = 10  # number of time series
    T = 100  # number of timesteps
    prediction_length = 24
    freq = "1H"
    custom_dataset = np.zeros(shape=(N, T))
    start = pd.Period(
        "01-01-2019", freq=freq
    )  # can be different for each time series
    train_ds = ListDataset(
        [
            {"target": x, "start": start}
            for x in custom_dataset[:, :-prediction_length]
        ],
        freq=freq,
    )
    # create a simple model
    estimator = SimpleFeedForwardEstimator(
        num_hidden_dimensions=[10],
        prediction_length=prediction_length,
        context_length=T,
        trainer=Trainer(
            ctx="cpu",
            epochs=1,
            learning_rate=1e-3,
            num_batches_per_epoch=1,
        ),
    )

    # train model
    predictor = estimator.train(train_ds)

    return predictor.prediction_net


@pytest.mark.parametrize(
    "n", [-2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 20]
)
@pytest.mark.parametrize("last_n_trigger", [True, False])
def test_NTA_V1(n: int, last_n_trigger: bool):
    model = initialize_model()
    params = model.collect_params()
    loss_list = [5, 4, 3, 2, 3, 3, 3, 3, 3, 3, 3]
    avg_strategy = NTA(
        epochs=len(loss_list),
        n=n,
        last_n_trigger=last_n_trigger,
        fallback_alpha=0.0,
    )
    for i, loss in enumerate(loss_list):
        for k, v in params.items():
            for arr in v.list_data():
                arr[:] = i
        avg_strategy.update_average_trigger(metric=loss)
        avg_strategy.apply(model)
    # nothing is cached yet, thus load_cached_model won't change anything
    # test cached model
    avg_strategy.load_cached_model(model)
    for k, v in params.items():
        for arr in v.list_data():
            # the last model should have 10 in all coordinates
            assert mx.nd.norm(arr - 10).asscalar() < 1e-30
    # test averaged model
    avg_strategy.load_averaged_model(model)
    len_limit = (len(loss_list) - 1) if last_n_trigger else 6
    if n <= 0 or n > len_limit:
        # average never happens, model is not changed
        for k, v in params.items():
            for arr in v.list_data():
                # the last model should have 10 in all coordinates
                assert mx.nd.norm(arr - 10).asscalar() < 1e-30
    else:
        for k, v in params.items():
            for arr in v.list_data():
                if last_n_trigger:
                    # takes the average once the loss increases, no matter what n is taken
                    # (the first n iterations are ignored)
                    val = 7 if n <= 4 else ((n + 10) / 2.0)
                else:
                    # takes the average on the last 7-n iterations
                    val = (4 + n + 10) / 2.0
                assert mx.nd.norm(arr - val).asscalar() < 1e-30
    # test cached model
    avg_strategy.load_cached_model(model)
    for k, v in params.items():
        for arr in v.list_data():
            # the last model should have 10 in all coordinates
            assert mx.nd.norm(arr - 10).asscalar() < 1e-30


@pytest.mark.parametrize("alpha", [0.0, 0.1, 0.2, 0.4, 0.6, 0.8, 1.0])
def test_Alpha_Suffix(alpha: float):
    model = initialize_model()
    params = model.collect_params()
    loss_list = [5, 4, 3, 2, 3, 3, 3, 3, 3, 3, 3]
    avg_strategy = Alpha_Suffix(epochs=len(loss_list), alpha=alpha)
    for i, loss in enumerate(loss_list):
        for k, v in params.items():
            for arr in v.list_data():
                arr[:] = i
        avg_strategy.update_average_trigger(epoch=i + 1)
        avg_strategy.apply(model)
    # nothing is cached yet, thus load_cached_model won't change anything
    # test cached model
    avg_strategy.load_cached_model(model)
    for k, v in params.items():
        for arr in v.list_data():
            # the last model should have 10 in all coordinates
            assert mx.nd.norm(arr - 10).asscalar() < 1e-30
    # test averaged model
    avg_strategy.load_averaged_model(model)
    n = max(int(math.ceil(len(loss_list) * (1 - alpha))), 1)
    if n > len(loss_list):
        # average never happens, model is not changed
        for k, v in params.items():
            for arr in v.list_data():
                # the last model should have 10 in all coordinates
                assert mx.nd.norm(arr - 10).asscalar() < 1e-30
    else:
        for k, v in params.items():
            for arr in v.list_data():
                val = (n + 9) / 2.0
                assert mx.nd.norm(arr - val).asscalar() < 1e-30
    # test cached model
    avg_strategy.load_cached_model(model)
    for k, v in params.items():
        for arr in v.list_data():
            # the last model should have 10 in all coordinates
            assert mx.nd.norm(arr - 10).asscalar() < 1e-30



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import tempfile
from pathlib import Path
from typing import cast, List
import numpy as np
from gluonts_customized.model.predictor import Predictor
from mxnet.gluon import nn
from tsbench.config import DatasetConfig, ModelConfig, TrainConfig
from tsbench.gluonts_customized.callbacks import (
    Callback,
    LearningRateScheduleCallback,
    ModelSaverCallback,
    ParameterCountCallback,
)
from .evaluate import FitResult
from .logging import log_metric


def fit_estimator(  # pylint: disable=too-many-statements
    config: ModelConfig,
    dataset: DatasetConfig,
    num_learning_rate_decays: int = 3,
    num_model_checkpoints: int = 5,
    validate: bool = False,
    verbose: bool = True,
) -> FitResult:
    """
    Fits the given estimator using the provided training dataset.

    Args:
        config: The configuration of the estimator to be fitted.
        dataset: The configuration of the dataset to be used for fitting.
        num_learning_rate_decays: The number of times the learning rate should be decayed.
        validate: Whether to use a validation dataset.
        choose_best: Whether the best model according to the validation loss within Hyperband
            intervals should be used.
        verbose: Whether to create multiple predictors and log associated information.

    Returns:
        The result from fitting, contains most notably the list of predictors fitted during
            training. Contains a single entry if the model is not trainable or `verbose` is set to
            false.
    """
    count_callback = ParameterCountCallback()
    callbacks: List[Callback] = [count_callback]

    milestones = []

    # We need to compute the full training time for the config on the dataset
    if isinstance(config, TrainConfig):
        training_time = config.training_fraction * dataset.max_training_time
    else:
        training_time = 0

    with tempfile.TemporaryDirectory() as tmp_dir:
        # If model is trainable, we need to create the callback
        saver_callback: ModelSaverCallback
        if isinstance(config, TrainConfig):
            hyperband_milestones = [
                training_time * (1 / 3) ** i
                for i in reversed(range(num_model_checkpoints))
            ]

            # Save at the first half of milestones as well as multiples of the middle milestone
            for i in range(len(hyperband_milestones) // 2):
                milestones += [hyperband_milestones[i]]

            pivot = hyperband_milestones[len(hyperband_milestones) // 2]
            milestones += np.arange(
                pivot, training_time + pivot / 2, pivot
            ).tolist()

            saver_callback = ModelSaverCallback(Path(tmp_dir), milestones)
            callbacks += [saver_callback]

            if num_learning_rate_decays > 0:
                learning_rate_callback = LearningRateScheduleCallback(
                    milestones=[
                        (training_time / (num_learning_rate_decays + 1)) * i
                        for i in range(1, num_learning_rate_decays + 1)
                    ],
                )
                callbacks += [learning_rate_callback]

        # Then, we can create the estimator
        meta = dataset.meta
        estimator = config.create_estimator(
            freq=meta.freq,
            prediction_length=cast(int, meta.prediction_length),
            time_features=dataset.has_time_features,
            training_time=training_time,
            validation_milestones=milestones if validate else [],
            callbacks=callbacks,
        )

        # Afterwards, we run the training. First, we need to optionally add validation data.
        train_kwargs = {}
        if isinstance(config, TrainConfig) and validate:
            train_kwargs["validation_data"] = dataset.data.val().gluonts()

        # Then, we can obtain the predictor
        train_data = dataset.data.train(validate).gluonts()
        predictor = estimator.train(train_data, **train_kwargs)

        # If the model is not trainable, we can return already, logging the recorded time
        if not isinstance(config, TrainConfig):
            if verbose:
                log_metric("num_model_parameters", 0)
                log_metric("num_gradient_updates", 0)
                log_metric("training_time", 0)
            return FitResult(config, [predictor], [0.0], 0)

        # Otherwise, we need to load all models that were stored by the callback
        predictors = []
        model_paths = []

        with tempfile.TemporaryDirectory() as model_dir:
            for i, params in enumerate(saver_callback.saved_parameters):  # type: ignore
                # Load the parameters
                saver_callback.network.load_parameters(params.absolute().as_posix())  # type: ignore

                # Create the predictor
                predictor = cast(TrainConfig, config).create_predictor(
                    estimator,
                    cast(nn.HybridBlock, saver_callback.network),  # type: ignore
                )
                # Serialize and deserialize to properly copy parameters
                path = Path(model_dir) / f"model_{i}"
                path.mkdir()

                predictor.serialize(path)
                model_paths.append(path)

                copied_predictor = Predictor.deserialize(model_paths[i])
                predictors.append(copied_predictor)

                # Log everything
                log_metric(
                    "num_model_parameters", count_callback.num_parameters
                )
                log_metric(
                    "num_gradient_updates",
                    saver_callback.num_gradient_updates[i],  # type: ignore
                )
                log_metric("training_time", saver_callback.training_times[i])  # type: ignore

        return FitResult(
            config,
            predictors,
            saver_callback.training_times,  # type: ignore
            count_callback.num_parameters,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from pathlib import Path
from typing import Dict, Optional, List, Iterator, Tuple

import pandas as pd

from gluonts_customized.core.component import validated
from gluonts_customized.dataset.common import Dataset
from gluonts_customized.dataset.util import forecast_start
from gluonts_customized.model.forecast import Forecast
from gluonts_customized.model.predictor import RepresentablePredictor
from gluonts_customized.time_feature import get_seasonality

# https://stackoverflow.com/questions/25329955/check-if-r-is-installed-from-python
from subprocess import Popen, PIPE

proc = Popen(["which", "R"], stdout=PIPE, stderr=PIPE)
R_IS_INSTALLED = proc.wait() == 0

try:
    import rpy2.robjects.packages as rpackages
    from rpy2 import rinterface, robjects
    from rpy2.rinterface_lib import callbacks
    from rpy2.rinterface_lib.embedded import RRuntimeError
except ImportError as e:
    rpy2_error_message = str(e)
    RPY2_IS_INSTALLED = False
else:
    RPY2_IS_INSTALLED = True

USAGE_MESSAGE = """
The RForecastPredictor is a thin wrapper for calling the R forecast package.
In order to use it you need to install R and rpy2. You also need to install \
specific R packages for univariate and/or hierarchical methods.

For univariate methods, install:
R -e 'install.packages(c("forecast", "nnfor"),\
repos="https://cloud.r-project.org")'

For hierarchical methods, install:
R -e 'install.packages(c("hts"), repos="https://cloud.r-project.org")'
"""


class RBasePredictor(RepresentablePredictor):
    """
    The `RBasePredictor` is a thin wrapper for calling R packages. In order to
    use it you need to install R and rpy2.

    Note that specific R packages need to be installed, depending
    on which wrapper one needs to run.
    See `RForecastPredictor` and `RHierarchicalForecastPredictor` to know
    which packages are needed.

    Parameters
    ----------
    freq
        The granularity of the time series (e.g. '1H')
    prediction_length
        Number of time points to be predicted.
    period
        The period to be used (this is called `frequency` in the R forecast
        package), result to a tentative reasonable default if not specified
        (for instance 24 for hourly freq '1H')
    trunc_length
        Maximum history length to feed to the model (some models become slow
        with very long series).
    r_file_prefix
        Prefix string of the R file(s) where our forecasting wrapper methods
        can be found.
        This is to avoid loading all R files potentially having different
        implementations of the same method, thereby making sure the
        expected R method is in fact used.
    """

    @validated()
    def __init__(
        self,
        freq: str,
        prediction_length: int,
        period: Optional[int] = None,
        trunc_length: Optional[int] = None,
        save_info: bool = False,
        r_file_prefix: str = "",
    ) -> None:
        super().__init__(prediction_length=prediction_length)

        if not R_IS_INSTALLED:
            raise ImportError("R is not Installed! \n " + USAGE_MESSAGE)

        if not RPY2_IS_INSTALLED:
            raise ImportError(rpy2_error_message + USAGE_MESSAGE)

        self._robjects = robjects
        self._rinterface = rinterface
        self._rcallbacks = callbacks
        self._rinterface.initr()
        self._rpackages = rpackages

        this_dir = Path(__file__).resolve().parent.absolute()
        this_dir = Path(f"{this_dir}".replace("\\", "/"))  # for windows
        r_files = this_dir.rglob(f"{r_file_prefix}*.R")

        for r_file in r_files:
            try:
                robjects.r(f'source("{r_file}")'.replace("\\", "\\\\"))
            except RRuntimeError as er:
                raise RRuntimeError(str(er) + USAGE_MESSAGE) from er

        self._stats_pkg = rpackages.importr("stats")

        self.prediction_length = prediction_length
        self.period = period if period is not None else get_seasonality(freq)
        self.trunc_length = trunc_length
        self.save_info = save_info

    def _get_r_forecast(self, data: Dict) -> Dict:
        """
        Get forecasts from an R method.

        Parameters
        ----------
        data
            Dictionary containing the `target` time series.

        Returns
        -------
        Dictionary
            Forecasts saved in a dictionary.
        """
        raise NotImplementedError()

    def _run_r_forecast(self, data: Dict) -> Tuple[Dict, List]:
        """
        Run an R forecast method.

        Parameters
        ----------
        data
            Dictionary containing the `target` time series.

        Returns
        -------
        Tuple[Dict, List]:
        """
        buf = []

        def save_to_buf(x):
            buf.append(x)

        def dont_save(x):
            pass

        f = save_to_buf if self.save_info else dont_save

        # save output from the R console in buf
        consolewrite_print_backup = self._rcallbacks.consolewrite_print
        consolewrite_warnerror_backup = self._rcallbacks.consolewrite_warnerror

        self._rcallbacks.consolewrite_print = f
        self._rcallbacks.consolewrite_warnerror = f

        forecast_dict = self._get_r_forecast(data=data)

        self._rcallbacks.consolewrite_print = consolewrite_print_backup
        self._rcallbacks.consolewrite_warnerror = consolewrite_warnerror_backup

        return forecast_dict, buf

    def _preprocess_data(self, data: Dict) -> Dict:
        """
        Preprocessing of target time series, e.g., truncating length or slicing
        bottom time series in case of hierarchical forecasting etc.

        Parameters
        ----------
        data
            Dictionary containing target time series.

        Returns
        -------
        Dict
        """
        raise NotImplementedError()

    def _warning_message(self) -> None:
        """
        Prints warning messages (once per whole dataset), e.g., if default
        parameters are overridden.
        """
        return

    def _forecast_dict_to_obj(
        self,
        forecast_dict: Dict,
        forecast_start_date: pd.Timestamp,
        item_id: Optional[str],
        info: Optional[Dict],
    ) -> Forecast:
        """
        Returns object of type `gluonts_customized.model.Forecast`.

        Parameters
        ----------
        forecast_dict
            Dictionary containing `samples` or `quantiles`.
        forecast_start_date
            Start date of the forecast.
        item_id
            Item identifier.
        info
            Additional information.

        Returns
        -------
        Forecast
            Sample based or quantile based forecasts.
        """
        raise NotImplementedError()

    def predict(
        self,
        dataset: Dataset,
        **kwargs,
    ) -> Iterator[Forecast]:
        self._warning_message()

        for data in dataset:
            data = self._preprocess_data(data=data)

            forecast_dict, console_output = self._run_r_forecast(data)

            info = (
                {"console_output": "\n".join(console_output)}
                if self.save_info
                else None
            )

            yield self._forecast_dict_to_obj(
                forecast_dict=forecast_dict,
                forecast_start_date=forecast_start(data),
                item_id=data.get("item_id", None),
                info=info,
            )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import io
from typing import Callable, List, Union

import pandas as pd
import numpy as np
import pytest

from gluonts_customized.dataset import pandas
from gluonts_customized.testutil.equality import assert_recursively_equal


@pytest.fixture(params=[pd.date_range, pd.period_range])
def my_series(request):
    idx = request.param("2021-01-01", freq="1D", periods=3)
    series = pd.Series(np.random.normal(size=3), index=idx)
    return series


@pytest.fixture()
def my_dataframe(my_series):
    return my_series.to_frame(name="target")


# elements will be called with my_series fixture, doesn't include long-format
all_formats = [
    lambda series: series,
    lambda series: series.to_frame("target"),
    lambda series: 3 * [series],
    lambda series: 3 * [series.to_frame("target")],
    lambda series: {i: series for i in ["A", "B", "C"]},
    lambda series: {i: series.to_frame("target") for i in ["A", "B", "C"]},
]


@pytest.fixture()
def long_dataframe():
    N, T = 2, 10
    df = pd.DataFrame(index=np.arange(N * T))
    df["time"] = (
        2 * pd.date_range("2021-01-01 00:00", freq="1H", periods=T).to_list()
    )
    df["target"] = np.random.normal(size=(N * T,))
    df["item"] = T * ["A"] + T * ["B"]
    df["stat_cat_1"] = T * [0] + T * [1]
    df["dyn_real_1"] = np.random.normal(size=(N * T,))
    return df


@pytest.fixture
def long_dataset(long_dataframe):  # initialized with dict
    return pandas.PandasDataset.from_long_dataframe(
        dataframe=long_dataframe,
        target="target",
        timestamp="time",
        item_id="item",
        freq="1H",
        feat_dynamic_real=["dyn_real_1"],
    )


@pytest.mark.parametrize("get_data", all_formats)
def test_PandasDataset_init_with_all_formats(get_data, my_series):
    dataset = pandas.PandasDataset(dataframes=get_data(my_series))
    assert len(dataset)
    for i in dataset:
        assert isinstance(i, dict)


def test_LongDataFrameDataset_iter(long_dataset):
    for i in long_dataset:
        assert isinstance(i, dict)
        assert "start" in i
        assert "target" in i
        assert "feat_dynamic_real" in i


def test_LongDataFrameDataset_len(long_dataset):
    assert len(long_dataset) == 2


def test_is_uniform_2H():
    timestamps = ["2021-01-01 00:00", "2021-01-01 02:00", "2021-01-01 04:00"]
    assert pandas.is_uniform(pd.DatetimeIndex(timestamps).to_period("2H"))


@pytest.mark.parametrize(
    "timestamps",
    [
        ["2021-01-01 00:00", "2021-01-01 02:00", "2021-01-01 02:00"],
        ["2021-01-01 00:00", "2021-01-01 02:00", "2021-01-01 03:00"],
        ["2021-01-01 04:00", "2021-01-01 02:00", "2021-01-01 00:00"],
    ],
)
def test_is_uniform_2H_fail(timestamps):
    assert not pandas.is_uniform(pd.DatetimeIndex(timestamps).to_period("2H"))


def test_infer_period(my_dataframe):
    ds = pandas.PandasDataset(my_dataframe, target="target", freq="1D")
    for entry in ds:
        assert entry["start"] == pd.Period("2021-01-01", freq="1D")


def test_infer_period2(my_dataframe):
    dfs = {"A": my_dataframe, "B": my_dataframe}
    ds = pandas.PandasDataset(dfs, target="target", freq="1D")
    for entry in ds:
        assert entry["start"] == pd.Period("2021-01-01", freq="1D")


def test_long_csv_3M():
    data = (
        "timestamp,item_id,target,static_feat\n"
        "2021-03,0,102,A\n"
        "2022-07,1,138,B\n"
        "2021-06,0,103,A\n"
        "2021-11,2,227,C\n"
        "2021-09,0,102,A\n"
        "2021-12,0,99,A\n"
        "2022-01,1,148,B\n"
        "2022-05,2,229,C\n"
        "2021-04,1,134,B\n"
        "2022-04,1,117,B\n"
        "2021-02,2,212,C\n"
        "2021-05,2,225,C\n"
        "2021-07,1,151,B\n"
        "2021-08,2,221,C\n"
        "2022-02,2,230,C\n"
        "2021-10,1,144,B\n"
    )

    expected_entries = [
        {
            "start": pd.Period("2021-03", freq="3M"),
            "item_id": 0,
            "feat_static_cat": np.array([0]),
            "target": np.array([102, 103, 102, 99]),
        },
        {
            "start": pd.Period("2021-04", freq="3M"),
            "item_id": 1,
            "feat_static_cat": np.array([1]),
            "target": np.array([134, 151, 144, 148, 117, 138]),
        },
        {
            "start": pd.Period("2021-02", freq="3M"),
            "item_id": 2,
            "feat_static_cat": np.array([2]),
            "target": np.array([212, 225, 221, 227, 230, 229]),
        },
    ]

    datasets = []

    with io.StringIO(data) as fp:
        df_long = pd.read_csv(fp)
        df_long["static_feat"] = df_long["static_feat"].astype("category")
        datasets.append(
            pandas.PandasDataset.from_long_dataframe(
                df_long,
                target="target",
                item_id="item_id",
                timestamp="timestamp",
                freq="3M",
                static_feature_columns=["static_feat"],
            )
        )

    with io.StringIO(data) as fp:
        df_long = pd.read_csv(fp, index_col="timestamp")
        df_long["static_feat"] = df_long["static_feat"].astype("category")
        datasets.append(
            pandas.PandasDataset.from_long_dataframe(
                df_long,
                target="target",
                item_id="item_id",
                freq="3M",
                static_feature_columns=["static_feat"],
            )
        )

    for ds in datasets:
        assert ds.static_cardinalities == np.array([3])
        assert isinstance(str(ds), str)
        for entry, expected_entry in zip(ds, expected_entries):
            assert entry["start"].freqstr == "3M"
            assert entry["start"] == expected_entry["start"]
            assert entry["item_id"] == expected_entry["item_id"]
            assert (
                entry["feat_static_cat"] == expected_entry["feat_static_cat"]
            )
            assert np.allclose(entry["target"], expected_entry["target"])


def _testcase_series(freq: str, index_type: Callable, dtype=np.float32):
    series = [
        pd.Series(
            np.arange(10, dtype=dtype),
            index_type("2021-01-01 00:00:00", periods=10, freq=freq),
        ),
        pd.Series(
            np.arange(20, dtype=dtype),
            index_type("2021-01-02 00:00:00", periods=20, freq=freq),
        ),
        pd.Series(
            np.arange(30, dtype=dtype),
            index_type("2021-01-03 00:00:00", periods=30, freq=freq),
        ),
    ]

    dataset = pandas.PandasDataset(series)

    expected_entries = [
        {
            "start": pd.Period(s.index[0], freq=freq),
            "target": s.values,
        }
        for s in series
    ]

    return dataset, expected_entries


def _testcase_dataframes_without_index(
    freq: str,
    target: Union[str, List[str]],
    feat_dynamic_real: List[str],
    dtype=np.float32,
):
    dataframes = [
        pd.DataFrame.from_dict(
            {
                "timestamp": pd.period_range(
                    "2021-01-01 00:00:00", periods=10, freq=freq
                )
                .map(str)
                .to_list(),
                "A": 1 + np.arange(10, dtype=dtype),
                "B": 2 + np.arange(10, dtype=dtype),
                "C": 3 + np.arange(10, dtype=dtype),
            }
        ),
        pd.DataFrame.from_dict(
            {
                "timestamp": pd.period_range(
                    "2021-01-02 00:00:00", periods=20, freq=freq
                )
                .map(str)
                .to_list(),
                "A": 1 + np.arange(20, dtype=dtype),
                "B": 2 + np.arange(20, dtype=dtype),
                "C": 3 + np.arange(20, dtype=dtype),
            }
        ),
        pd.DataFrame.from_dict(
            {
                "timestamp": pd.period_range(
                    "2021-01-03 00:00:00", periods=30, freq=freq
                )
                .map(str)
                .to_list(),
                "A": 1 + np.arange(30, dtype=dtype),
                "B": 2 + np.arange(30, dtype=dtype),
                "C": 3 + np.arange(30, dtype=dtype),
            }
        ),
    ]

    dataset = pandas.PandasDataset(
        dataframes,
        timestamp="timestamp",
        freq=freq,
        target=target,
        feat_dynamic_real=feat_dynamic_real,
    )

    expected_entries = [
        {
            "start": pd.Period(df["timestamp"][0], freq=freq),
            "target": df[target].values.transpose(),
            "feat_dynamic_real": df[feat_dynamic_real].values.transpose(),
        }
        for df in dataframes
    ]

    return dataset, expected_entries


def _testcase_dataframes_with_index(
    freq: str,
    index_type: Callable,
    target: Union[str, List[str]],
    feat_dynamic_real: List[str],
    dtype=np.float32,
):
    dataframes = [
        pd.DataFrame.from_dict(
            {
                "timestamp": index_type(
                    "2021-01-01 00:00:00", periods=10, freq=freq
                ),
                "A": 1 + np.arange(10, dtype=dtype),
                "B": 2 + np.arange(10, dtype=dtype),
                "C": 3 + np.arange(10, dtype=dtype),
            }
        ).set_index("timestamp"),
        pd.DataFrame.from_dict(
            {
                "timestamp": index_type(
                    "2021-01-02 00:00:00", periods=20, freq=freq
                ),
                "A": 1 + np.arange(20, dtype=dtype),
                "B": 2 + np.arange(20, dtype=dtype),
                "C": 3 + np.arange(20, dtype=dtype),
            }
        ).set_index("timestamp"),
        pd.DataFrame.from_dict(
            {
                "timestamp": index_type(
                    "2021-01-03 00:00:00", periods=30, freq=freq
                ),
                "A": 1 + np.arange(30, dtype=dtype),
                "B": 2 + np.arange(30, dtype=dtype),
                "C": 3 + np.arange(30, dtype=dtype),
            }
        ).set_index("timestamp"),
    ]

    print(type(dataframes[0].index))

    dataset = pandas.PandasDataset(
        dataframes,
        target=target,
        feat_dynamic_real=feat_dynamic_real,
    )

    expected_entries = [
        {
            "start": pd.Period(df.index[0], freq=freq),
            "target": df[target].values.transpose(),
            "feat_dynamic_real": df[feat_dynamic_real].values.transpose(),
        }
        for df in dataframes
    ]

    return dataset, expected_entries


@pytest.mark.parametrize(
    "dataset, expected_entries",
    [
        _testcase_series(freq="D", index_type=pd.period_range),
        _testcase_series(freq="H", index_type=pd.date_range),
        _testcase_dataframes_without_index(
            freq="D",
            target="A",
            feat_dynamic_real=["B", "C"],
        ),
        _testcase_dataframes_without_index(
            freq="H",
            target=["A", "B"],
            feat_dynamic_real=["C"],
        ),
        _testcase_dataframes_with_index(
            freq="D",
            index_type=pd.period_range,
            target="A",
            feat_dynamic_real=["B", "C"],
        ),
        _testcase_dataframes_with_index(
            freq="H",
            index_type=pd.date_range,
            target=["A", "B"],
            feat_dynamic_real=["C"],
        ),
    ],
)
def test_pandas_dataset_cases(dataset, expected_entries):
    for entry, expected_entry in zip(dataset, expected_entries):
        assert_recursively_equal(entry, expected_entry)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import logging
import shutil
from functools import partial
from pathlib import Path
from typing import Any, Dict, Optional

from gluonts_customized.dataset import DatasetWriter
from gluonts_customized.dataset.artificial import ConstantDataset
from gluonts_customized.dataset.common import TrainDatasets, load_datasets
from gluonts_customized.dataset.jsonl import JsonLinesWriter

from ._artificial import generate_artificial_dataset
from ._airpassengers import generate_airpassengers_dataset
from ._ercot import generate_ercot_dataset
from ._ett_small import generate_ett_small_dataset
from ._gp_copula_2019 import generate_gp_copula_dataset
from ._lstnet import generate_lstnet_dataset
from ._m3 import generate_m3_dataset
from ._m4 import generate_m4_dataset
from ._m5 import generate_m5_dataset
from ._tsf_datasets import generate_forecasting_dataset
from ._uber_tlc import generate_uber_dataset


def get_download_path() -> Path:
    """

    Returns
    -------
    Path
        default path to download datasets or models of gluon-ts.
        The path is $HOME/.gluonts/
    """
    return Path.home() / ".gluonts"


dataset_recipes = {
    # each recipe generates a dataset given a path
    "constant": partial(
        generate_artificial_dataset, dataset=ConstantDataset()
    ),
    "exchange_rate": partial(
        generate_lstnet_dataset, dataset_name="exchange_rate"
    ),
    "solar-energy": partial(
        generate_lstnet_dataset, dataset_name="solar-energy"
    ),
    "electricity": partial(
        generate_lstnet_dataset, dataset_name="electricity"
    ),
    "traffic": partial(generate_lstnet_dataset, dataset_name="traffic"),
    "exchange_rate_nips": partial(
        generate_gp_copula_dataset, dataset_name="exchange_rate_nips"
    ),
    "electricity_nips": partial(
        generate_gp_copula_dataset, dataset_name="electricity_nips"
    ),
    "traffic_nips": partial(
        generate_gp_copula_dataset, dataset_name="traffic_nips"
    ),
    "solar_nips": partial(
        generate_gp_copula_dataset, dataset_name="solar_nips"
    ),
    "wiki2000_nips": partial(
        generate_gp_copula_dataset, dataset_name="wiki2000_nips"
    ),
    "wiki-rolling_nips": partial(
        generate_gp_copula_dataset, dataset_name="wiki-rolling_nips"
    ),
    "taxi_30min": partial(
        generate_gp_copula_dataset, dataset_name="taxi_30min"
    ),
    "kaggle_web_traffic_with_missing": partial(
        generate_forecasting_dataset,
        dataset_name="kaggle_web_traffic_with_missing",
    ),
    "kaggle_web_traffic_without_missing": partial(
        generate_forecasting_dataset,
        dataset_name="kaggle_web_traffic_without_missing",
    ),
    "kaggle_web_traffic_weekly": partial(
        generate_forecasting_dataset,
        dataset_name="kaggle_web_traffic_weekly",
    ),
    "m1_yearly": partial(
        generate_forecasting_dataset, dataset_name="m1_yearly"
    ),
    "m1_quarterly": partial(
        generate_forecasting_dataset, dataset_name="m1_quarterly"
    ),
    "m1_monthly": partial(
        generate_forecasting_dataset, dataset_name="m1_monthly"
    ),
    "nn5_daily_with_missing": partial(
        generate_forecasting_dataset, dataset_name="nn5_daily_with_missing"
    ),
    "nn5_daily_without_missing": partial(
        generate_forecasting_dataset,
        dataset_name="nn5_daily_without_missing",
    ),
    "nn5_weekly": partial(
        generate_forecasting_dataset, dataset_name="nn5_weekly"
    ),
    "tourism_monthly": partial(
        generate_forecasting_dataset, dataset_name="tourism_monthly"
    ),
    "tourism_quarterly": partial(
        generate_forecasting_dataset, dataset_name="tourism_quarterly"
    ),
    "tourism_yearly": partial(
        generate_forecasting_dataset, dataset_name="tourism_yearly"
    ),
    "cif_2016": partial(generate_forecasting_dataset, dataset_name="cif_2016"),
    "london_smart_meters_without_missing": partial(
        generate_forecasting_dataset,
        dataset_name="london_smart_meters_without_missing",
    ),
    "wind_farms_without_missing": partial(
        generate_forecasting_dataset,
        dataset_name="wind_farms_without_missing",
    ),
    "car_parts_without_missing": partial(
        generate_forecasting_dataset,
        dataset_name="car_parts_without_missing",
    ),
    "dominick": partial(generate_forecasting_dataset, dataset_name="dominick"),
    "fred_md": partial(generate_forecasting_dataset, dataset_name="fred_md"),
    "pedestrian_counts": partial(
        generate_forecasting_dataset, dataset_name="pedestrian_counts"
    ),
    "hospital": partial(generate_forecasting_dataset, dataset_name="hospital"),
    "covid_deaths": partial(
        generate_forecasting_dataset, dataset_name="covid_deaths"
    ),
    "kdd_cup_2018_without_missing": partial(
        generate_forecasting_dataset,
        dataset_name="kdd_cup_2018_without_missing",
    ),
    "weather": partial(generate_forecasting_dataset, dataset_name="weather"),
    "m3_monthly": partial(generate_m3_dataset, m3_freq="monthly"),
    "m3_quarterly": partial(generate_m3_dataset, m3_freq="quarterly"),
    "m3_yearly": partial(generate_m3_dataset, m3_freq="yearly"),
    "m3_other": partial(generate_m3_dataset, m3_freq="other"),
    "m4_hourly": partial(
        generate_m4_dataset,
        m4_freq="Hourly",
        pandas_freq="H",
        prediction_length=48,
    ),
    "m4_daily": partial(
        generate_m4_dataset,
        m4_freq="Daily",
        pandas_freq="D",
        prediction_length=14,
    ),
    "m4_weekly": partial(
        generate_m4_dataset,
        m4_freq="Weekly",
        pandas_freq="W",
        prediction_length=13,
    ),
    "m4_monthly": partial(
        generate_m4_dataset,
        m4_freq="Monthly",
        pandas_freq="M",
        prediction_length=18,
    ),
    "m4_quarterly": partial(
        generate_m4_dataset,
        m4_freq="Quarterly",
        pandas_freq="Q",
        prediction_length=8,
    ),
    "m4_yearly": partial(
        generate_m4_dataset,
        m4_freq="Yearly",
        pandas_freq="Y",
        prediction_length=6,
    ),
    "m5": partial(
        generate_m5_dataset,
        pandas_freq="D",
        prediction_length=28,
        m5_file_path=get_download_path() / "m5",
    ),
    "uber_tlc_daily": partial(
        generate_uber_dataset, uber_freq="Daily", prediction_length=7
    ),
    "uber_tlc_hourly": partial(
        generate_uber_dataset, uber_freq="Hourly", prediction_length=24
    ),
    "airpassengers": partial(generate_airpassengers_dataset),
    "australian_electricity_demand": partial(
        generate_forecasting_dataset,
        dataset_name="australian_electricity_demand",
    ),
    "electricity_hourly": partial(
        generate_forecasting_dataset,
        dataset_name="electricity_hourly",
    ),
    "electricity_weekly": partial(
        generate_forecasting_dataset,
        dataset_name="electricity_weekly",
    ),
    "rideshare_without_missing": partial(
        generate_forecasting_dataset,
        dataset_name="rideshare_without_missing",
    ),
    "saugeenday": partial(
        generate_forecasting_dataset,
        dataset_name="saugeenday",
    ),
    "solar_10_minutes": partial(
        generate_forecasting_dataset,
        dataset_name="solar_10_minutes",
    ),
    "solar_weekly": partial(
        generate_forecasting_dataset,
        dataset_name="solar_weekly",
    ),
    "sunspot_without_missing": partial(
        generate_forecasting_dataset,
        dataset_name="sunspot_without_missing",
    ),
    "temperature_rain_without_missing": partial(
        generate_forecasting_dataset,
        dataset_name="temperature_rain_without_missing",
    ),
    "vehicle_trips_without_missing": partial(
        generate_forecasting_dataset,
        dataset_name="vehicle_trips_without_missing",
    ),
    "ercot": partial(generate_ercot_dataset),
    "ett_small_15min": partial(
        generate_ett_small_dataset,
        base_file_name="ETTm",
        freq="15min",
        prediction_length=24,
    ),
    "ett_small_1h": partial(
        generate_ett_small_dataset,
        base_file_name="ETTh",
        freq="1h",
        prediction_length=24,
    ),
}


dataset_names = list(dataset_recipes.keys())

default_dataset_path = get_download_path() / "datasets"

default_dataset_writer = JsonLinesWriter()


def materialize_dataset(
    dataset_name: str,
    path: Path = default_dataset_path,
    regenerate: bool = False,
    dataset_writer: DatasetWriter = default_dataset_writer,
    prediction_length: Optional[int] = None,
) -> Path:
    """
    Ensures that the dataset is materialized under the `path / dataset_name`
    path.

    Parameters
    ----------
    dataset_name
        Name of the dataset, for instance "m4_hourly".
    regenerate
        Whether to regenerate the dataset even if a local file is present.
        If this flag is False and the file is present, the dataset will not
        be downloaded again.
    path
        Where the dataset should be saved.
    prediction_length
        The prediction length to be used for the dataset. If None, the default
        prediction length will be used. The prediction length might not be
        available for all datasets.

    Returns
    -------
        The path where the dataset is materialized
    """
    assert dataset_name in dataset_recipes.keys(), (
        f"{dataset_name} is not present, please choose one from "
        f"{sorted(dataset_recipes)}."
    )

    path.mkdir(parents=True, exist_ok=True)
    dataset_path = path / dataset_name

    dataset_recipe = dataset_recipes[dataset_name]

    if not dataset_path.exists() or regenerate:
        logging.info(f"downloading and processing {dataset_name}")
        if dataset_path.exists():
            # If regenerating, we need to remove the directory contents
            shutil.rmtree(dataset_path)
            dataset_path.mkdir()

        # Optionally pass prediction length to not override any non-None
        # defaults (e.g. for M4)
        kwargs: Dict[str, Any] = {"dataset_path": dataset_path}
        if prediction_length is not None:
            kwargs["prediction_length"] = prediction_length
        kwargs["dataset_writer"] = dataset_writer
        dataset_recipe(**kwargs)
    else:
        logging.info(
            f"using dataset already processed in path {dataset_path}."
        )

    return dataset_path


def get_dataset(
    dataset_name: str,
    path: Path = default_dataset_path,
    regenerate: bool = False,
    dataset_writer: DatasetWriter = default_dataset_writer,
    prediction_length: Optional[int] = None,
) -> TrainDatasets:
    """
    Get a repository dataset.

    The datasets that can be obtained through this function have been used
    with different processing over time by several papers (e.g., [SFG17]_,
    [LCY+18]_, and [YRD15]_) or are obtained through the `Monash Time Series
    Forecasting Repository <https://forecastingdata.org/>`_.

    Parameters
    ----------
    dataset_name
        Name of the dataset, for instance "m4_hourly".
    regenerate
        Whether to regenerate the dataset even if a local file is present.
        If this flag is False and the file is present, the dataset will not
        be downloaded again.
    path
        Where the dataset should be saved.
    prediction_length
        The prediction length to be used for the dataset. If None, the default
        prediction length will be used. If the dataset is already materialized,
        setting this option to a different value does not have an effect.
        Make sure to set `regenerate=True` in this case. Note that some
        datasets from the Monash Time Series Forecasting Repository do not
        actually have a default prediction length -- the default then depends
        on the frequency of the data:
        - Minutely data --> prediction length of 60 (one hour)
        - Hourly data --> prediction length of 48 (two days)
        - Daily data --> prediction length of 30 (one month)
        - Weekly data --> prediction length of 8 (two months)
        - Monthly data --> prediction length of 12 (one year)
        - Yearly data --> prediction length of 4 (four years)

    Returns
    -------
        Dataset obtained by either downloading or reloading from local file.
    """
    dataset_path = materialize_dataset(
        dataset_name,
        path,
        regenerate,
        dataset_writer,
        prediction_length,
    )

    return load_datasets(
        metadata=dataset_path,
        train=dataset_path / "train",
        test=dataset_path / "test",
    )


if __name__ == "__main__":
    for dataset in dataset_names:
        print(f"generate {dataset}")
        ds = get_dataset(dataset, regenerate=True)
        print(ds.metadata)
        print(len(ds.train))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from itertools import tee
from typing import List

import numpy as np

from gluonts_customized.dataset.common import Dataset
from gluonts_customized.evaluation import Evaluator
from gluonts_customized.model.predictor import Predictor

from gluonts_customized.nursery.temporal_hierarchical_forecasting.utils.common import (
    TEMPORAL_HIERARCHIES,
)
from gluonts_customized.nursery.temporal_hierarchical_forecasting.utils.utils import (
    get_ts_at_all_levels,
    unpack_forecasts,
)
from gluonts_customized.nursery.temporal_hierarchical_forecasting.utils.gluonts_helper import (
    make_predictions,
    to_dataframe_it,
    truncate_target,
)


def evaluate_predictor(
    predictor: Predictor,
    test_dataset: Dataset,
    freq: str,
    metrics: List[str] = ["mean_wQuantileLoss"],
    evaluate_all_levels: bool = False,
):
    pred_input_it = truncate_target(
        test_dataset,
        prediction_length=predictor.prediction_length,
    )

    forecast_it = make_predictions(
        pred_input=pred_input_it,
        predictor=predictor,
        forecasts_at_all_levels=evaluate_all_levels,
    )
    test_ts_it = to_dataframe_it(test_dataset)

    evaluator = Evaluator(quantiles=(np.arange(20) / 20.0)[1:], num_workers=0)
    if not evaluate_all_levels:
        # Evaluate only the bottom level.
        agg_metrics, _ = evaluator(
            ts_iterator=test_ts_it,
            fcst_iterator=forecast_it,
        )
        metrics_to_return = {
            metric_str: agg_metrics[metric_str] for metric_str in metrics
        }
        return metrics_to_return

    # Evaluate all levels.
    metrics_to_return = evaluate_forecasts_at_all_levels(
        forecast_at_all_levels_it=forecast_it,
        test_ts_at_all_levels_it=get_ts_at_all_levels(
            ts_it=test_ts_it,
            temporal_hierarchy=TEMPORAL_HIERARCHIES[freq],
            prediction_length=predictor.prediction_length,
            target_temporal_hierarchy=TEMPORAL_HIERARCHIES[freq],
        ),
        temporal_hierarchy=TEMPORAL_HIERARCHIES[freq],
        evaluator=evaluator,
        metrics=metrics,
    )
    return metrics_to_return


def evaluate_forecasts_at_all_levels(
    forecast_at_all_levels_it,
    test_ts_at_all_levels_it,
    temporal_hierarchy,
    evaluator,
    metrics: List[str] = ["mean_wQuantileLoss"],
):
    # Order of forecasts obtained e.g., in case of the hierarchy: [6M, 2M, 1M]
    # (ts_1_6M, ts_1_2M, ts_1_1M, ts_2_6M, ts_2_2M, ts_2_1M, ...)
    forecast_at_all_levels_unpacked_it = unpack_forecasts(
        forecast_at_all_levels_it=forecast_at_all_levels_it,
        temporal_hierarchy=temporal_hierarchy,
        target_temporal_hierarchy=temporal_hierarchy,
    )

    num_levels = len(temporal_hierarchy.agg_multiples)

    # In one go, we can get item metrics for time series that have the same frequency.
    # So we create `num_levels` copies of the iterator and obtain the item metrics
    # for each level independently.
    forecast_at_all_levels_unpacked_it_set = tee(
        forecast_at_all_levels_unpacked_it,
        num_levels,
    )
    test_ts_at_all_levels_it_set = tee(test_ts_at_all_levels_it, num_levels)

    # Since forecasts for all granularities are in the same iterable,
    # we need a way to iterate through forecasts skipping some elements.
    def skip_iter(it, num_skips: int, offset: int):
        for _ in range(offset):
            next(it)
        for item in it:
            for _ in range(num_skips):
                next(it, None)  # None: in case the `it` is already exhausted.
            yield item

    metrics_to_return = {}
    for level in range(num_levels):
        agg_metrics_level, metrics_per_ts_level = evaluator(
            ts_iterator=skip_iter(
                test_ts_at_all_levels_it_set[level],
                num_skips=num_levels - 1,
                offset=level,
            ),
            fcst_iterator=skip_iter(
                forecast_at_all_levels_unpacked_it_set[level],
                num_skips=num_levels - 1,
                offset=level,
            ),
        )

        for metric_name in metrics:
            metrics_to_return[
                f"Freq_{temporal_hierarchy.freq_strs[level]}_{metric_name}"
            ] = agg_metrics_level[metric_name]

    return metrics_to_return



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import os
import tarfile
from distutils.util import strtobool
from functools import partial
from typing import Dict, Optional, Union

from toolz import valmap

from gluonts_customized.dataset.common import Dataset, FileDataset, MetaData
from gluonts_customized.model import Predictor
from gluonts_customized.util import safe_extractall

from . import sagemaker


class TrainEnv(sagemaker.TrainEnv):
    def __init__(self, *args, **kwargs):
        sagemaker.TrainEnv.__init__(self, *args, **kwargs)
        self.datasets = self._load()

    def _load(self) -> Dict[str, Union[Dataset, Predictor]]:
        if "metadata" in self.channels:
            path = self.channels.pop("metadata")
            self.hyperparameters["freq"] = MetaData.parse_file(
                path / "metadata.json"
            ).freq

        model: Optional[Predictor] = None
        if "model" in self.channels:
            path = self.channels.pop("model")
            with tarfile.open(path / "model.tar.gz") as targz:
                safe_extractall(targz, path)
            model = Predictor.deserialize(path)

        file_dataset = partial(FileDataset, freq=self.hyperparameters["freq"])

        datasets = valmap(file_dataset, self.channels)
        if self._listify_dataset():
            datasets = valmap(list, datasets)
        if model is not None:
            datasets["model"] = model

        return datasets

    def _listify_dataset(self):
        return strtobool(self.hyperparameters.get("listify_dataset", "no"))


class ServeEnv(sagemaker.ServeEnv):
    def __init__(self, *args, **kwargs):
        sagemaker.ServeEnv.__init__(self, *args, **kwargs)

        if self.sagemaker_batch:
            from .serve.app import ForecastConfig

            self.batch_config = ForecastConfig.parse_raw(
                os.environ["INFERENCE_CONFIG"]
            )
        else:
            self.batch_config = None



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import sys
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from gluonts_customized.dataset.common import FileDataset
from gluonts_customized.dataset.arrow import ArrowWriter, ParquetWriter
from gluonts_customized.dataset.jsonl import JsonLinesWriter
from gluonts_customized.dataset.repository import get_dataset


@pytest.mark.skipif(sys.version_info < (3, 7), reason="Requires PyArrow v8.")
@pytest.mark.parametrize(
    "writer", [ArrowWriter(), ParquetWriter(), JsonLinesWriter()]
)
def test_dataset_writer(writer):
    dataset = get_dataset("constant")

    with TemporaryDirectory() as temp_dir:
        writer.write_to_folder(dataset.train, Path(temp_dir))

        loaded = FileDataset(Path(temp_dir), freq="h")
        assert len(dataset.train) == len(loaded)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import abc
from dataclasses import dataclass
from typing import Callable, Iterable, Iterator, List

from gluonts_customized.core.component import equals, validated
from gluonts_customized.dataset.common import DataEntry, Dataset
from gluonts_customized.env import env


class Transformation(metaclass=abc.ABCMeta):
    """
    Base class for all Transformations.

    A Transformation processes works on a stream (iterator) of dictionaries.
    """

    @abc.abstractmethod
    def __call__(
        self, data_it: Iterable[DataEntry], is_train: bool
    ) -> Iterable[DataEntry]:
        pass

    def chain(self, other: "Transformation") -> "Chain":
        return Chain([self, other])

    def __add__(self, other: "Transformation") -> "Chain":
        return self.chain(other)

    def apply(
        self, dataset: Dataset, is_train: bool = True
    ) -> "TransformedDataset":
        return TransformedDataset(dataset, self, is_train=is_train)


@dataclass
class Chain(Transformation):
    """
    Chain multiple transformations together.
    """

    transformations: List[Transformation]

    def __post_init__(self) -> None:
        transformations = []

        for transformation in self.transformations:
            if isinstance(transformation, Identity):
                continue
            elif isinstance(transformation, Chain):
                transformations.extend(transformation.transformations)
            else:
                assert isinstance(transformation, Transformation)
                transformations.append(transformation)

        self.transformations = transformations
        self.__init_passed_kwargs__ = {"transformations": transformations}

    def __call__(
        self, data_it: Iterable[DataEntry], is_train: bool
    ) -> Iterable[DataEntry]:
        tmp = data_it
        for t in self.transformations:
            tmp = t(tmp, is_train)
        return tmp


class TransformedDataset(Dataset):
    """
    A dataset that corresponds to applying a list of transformations to each
    element in the base_dataset. This only supports SimpleTransformations,
    which do the same thing at prediction and training time.

    Parameters
    ----------
    base_dataset
        Dataset to transform
    transformations
        List of transformations to apply
    """

    def __init__(
        self,
        base_dataset: Dataset,
        transformation: Transformation,
        is_train=True,
    ) -> None:
        self.base_dataset = base_dataset
        self.transformation = transformation
        self.is_train = is_train

    def __len__(self):
        # NOTE this is unsafe when transformations are run with is_train = True
        # since some transformations may not be deterministic
        # (instance splitter)
        return sum(1 for _ in self)

    def __iter__(self) -> Iterator[DataEntry]:
        yield from self.transformation(
            self.base_dataset, is_train=self.is_train
        )


class Identity(Transformation):
    def __call__(
        self, data_it: Iterable[DataEntry], is_train: bool
    ) -> Iterable[DataEntry]:
        return data_it


class MapTransformation(Transformation):
    """
    Base class for Transformations that returns exactly one result per input in
    the stream.
    """

    def __call__(
        self, data_it: Iterable[DataEntry], is_train: bool
    ) -> Iterator:
        for data_entry in data_it:
            try:
                yield self.map_transform(data_entry.copy(), is_train)
            except Exception as e:
                raise e

    @abc.abstractmethod
    def map_transform(self, data: DataEntry, is_train: bool) -> DataEntry:
        pass


class SimpleTransformation(MapTransformation):
    """
    Element wise transformations that are the same in train and test mode.
    """

    def map_transform(self, data: DataEntry, is_train: bool) -> DataEntry:
        return self.transform(data)

    @abc.abstractmethod
    def transform(self, data: DataEntry) -> DataEntry:
        pass


class AdhocTransform(SimpleTransformation):
    """
    Applies a function as a transformation This is called ad-hoc, because it is
    not serializable.

    It is OK to use this for experiments and outside of a model pipeline that
    needs to be serialized.
    """

    def __init__(self, func: Callable[[DataEntry], DataEntry]) -> None:
        self.func = func

    def transform(self, data: DataEntry) -> DataEntry:
        return self.func(data.copy())


class FlatMapTransformation(Transformation):
    """
    Transformations that yield zero or more results per input, but do not
    combine elements from the input stream.
    """

    @validated()
    def __init__(self):
        self.max_idle_transforms = env.max_idle_transforms

    def __call__(
        self, data_it: Iterable[DataEntry], is_train: bool
    ) -> Iterator:
        num_idle_transforms = 0
        for data_entry in data_it:
            num_idle_transforms += 1
            for result in self.flatmap_transform(data_entry.copy(), is_train):
                num_idle_transforms = 0
                yield result

            if (
                # negative values disable the check
                self.max_idle_transforms > 0
                and num_idle_transforms > self.max_idle_transforms
            ):
                raise Exception(
                    "Reached maximum number of idle transformation"
                    " calls.\nThis means the transformation looped over"
                    f" {self.max_idle_transforms} inputs without returning any"
                    " output.\nThis occurred in the following"
                    f" transformation:\n{self}"
                )

    @abc.abstractmethod
    def flatmap_transform(
        self, data: DataEntry, is_train: bool
    ) -> Iterator[DataEntry]:
        pass


class FilterTransformation(FlatMapTransformation):
    def __init__(self, condition: Callable[[DataEntry], bool]) -> None:
        super().__init__()
        self.condition = condition

    def flatmap_transform(
        self, data: DataEntry, is_train: bool
    ) -> Iterator[DataEntry]:
        if self.condition(data):
            yield data


# The __init__ in FilterTransformation is not validated but the __init__ in the
# parent class (FlatMapTransformation) is. So now the code
# (equals_default_impl) validate the arguments in FilterTransformation, which
# is an empty dict for all the FilterTransformation. We can not make __init__
# FilterTransformation as validated as we may use lambda function as the
# argument
@equals.register(FilterTransformation)
def equals_filter_transformation(
    this: FilterTransformation, that: FilterTransformation
):
    return this.condition.__code__.co_code == that.condition.__code__.co_code



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict, List, Optional, Tuple, cast

import torch
from torch.distributions import AffineTransform, TransformedDistribution
from torch.distributions.normal import Normal

from gluonts_customized.core.component import validated
from gluonts_customized.itertools import prod
from gluonts_customized.torch.distributions import DistributionOutput
from gluonts_customized.torch.model.mqf2.module import PICNN


class MQF2Distribution(torch.distributions.Distribution):
    r"""
    Distribution class for the model MQF2 proposed in the paper ``Multivariate
    Quantile Function Forecaster`` by Kan, Aubet, Januschowski, Park, Benidis,
    Ruthotto, Gasthaus.

    Parameters
    ----------
    picnn
        A SequentialNet instance of a
        partially input convex neural network (picnn)
    hidden_state
        hidden_state obtained by unrolling the RNN encoder
        shape = (batch_size, context_length, hidden_size) in training
        shape = (batch_size, hidden_size) in inference
    prediction_length
        Length of the prediction horizon
    is_energy_score
        If True, use energy score as objective function
        otherwise use maximum likelihood as
        objective function (normalizing flows)
    es_num_samples
        Number of samples drawn to approximate the energy score
    beta
        Hyperparameter of the energy score (power of the two terms)
    threshold_input
        Clamping threshold of the (scaled) input when maximum
        likelihood is used as objective function
        this is used to make the forecaster more robust
        to outliers in training samples
    validate_args
        Sets whether validation is enabled or disabled
        For more details, refer to the descriptions in
        torch.distributions.distribution.Distribution
    """

    def __init__(
        self,
        picnn: PICNN,
        hidden_state: torch.Tensor,
        prediction_length: int,
        is_energy_score: bool = True,
        es_num_samples: int = 50,
        beta: float = 1.0,
        threshold_input: float = 100.0,
        validate_args: bool = False,
    ) -> None:
        self.picnn = picnn
        self.hidden_state = hidden_state
        self.prediction_length = prediction_length
        self.is_energy_score = is_energy_score
        self.es_num_samples = es_num_samples
        self.beta = beta
        self.threshold_input = threshold_input

        super().__init__(
            batch_shape=self.batch_shape, validate_args=validate_args
        )

        self.context_length = (
            self.hidden_state.shape[-2]
            if len(self.hidden_state.shape) > 2
            else 1
        )
        self.numel_batch = prod(self.batch_shape)

        # mean zero and std one
        mu = torch.tensor(
            0, dtype=hidden_state.dtype, device=hidden_state.device
        )
        sigma = torch.ones_like(mu)
        self.standard_normal = Normal(mu, sigma)

    def stack_sliding_view(self, z: torch.Tensor) -> torch.Tensor:
        """
        Auxiliary function for loss computation.

        Unfolds the observations by sliding a window of size prediction_length
        over the observations z
        Then, reshapes the observations into a 2-dimensional tensor for
        further computation

        Parameters
        ----------
        z
            A batch of time series with shape
            (batch_size, context_length + prediction_length - 1)

        Returns
        -------
        Tensor
            Unfolded time series with shape
            (batch_size * context_length, prediction_length)
        """

        z = z.unfold(dimension=-1, size=self.prediction_length, step=1)
        z = z.reshape(-1, z.shape[-1])

        return z

    def loss(self, z: torch.Tensor) -> torch.Tensor:
        if self.is_energy_score:
            return self.energy_score(z)
        else:
            return -self.log_prob(z)

    def log_prob(self, z: torch.Tensor) -> torch.Tensor:
        """
        Computes the log likelihood  log(g(z)) + logdet(dg(z)/dz), where g is
        the gradient of the picnn.

        Parameters
        ----------
        z
            A batch of time series with shape
            (batch_size, context_length + prediciton_length - 1)

        Returns
        -------
        loss
            Tesnor of shape (batch_size * context_length,)
        """

        z = torch.clamp(z, min=-self.threshold_input, max=self.threshold_input)
        z = self.stack_sliding_view(z)

        loss = self.picnn.logp(
            z, self.hidden_state.reshape(-1, self.hidden_state.shape[-1])
        )

        return loss

    def energy_score(self, z: torch.Tensor) -> torch.Tensor:
        """
        Computes the (approximated) energy score sum_i ES(g,z_i), where
        ES(g,z_i) =

        -1/(2*es_num_samples^2) * sum_{w,w'} ||w-w'||_2^beta
        + 1/es_num_samples * sum_{w''} ||w''-z_i||_2^beta,
        w's are samples drawn from the
        quantile function g(., h_i) (gradient of picnn),
        h_i is the hidden state associated with z_i,
        and es_num_samples is the number of samples drawn
        for each of w, w', w'' in energy score approximation

        Parameters
        ----------
        z
            A batch of time series with shape
            (batch_size, context_length + prediction_length - 1)

        Returns
        -------
        loss
            Tensor of shape (batch_size * context_length,)
        """

        es_num_samples = self.es_num_samples
        beta = self.beta

        z = self.stack_sliding_view(z)
        reshaped_hidden_state = self.hidden_state.reshape(
            -1, self.hidden_state.shape[-1]
        )

        loss = self.picnn.energy_score(
            z, reshaped_hidden_state, es_num_samples=es_num_samples, beta=beta
        )

        return loss

    def rsample(self, sample_shape: torch.Size = torch.Size()) -> torch.Tensor:
        """
        Generates the sample paths.

        Parameters
        ----------
        sample_shape
            Shape of the samples

        Returns
        -------
        sample_paths
            Tesnor of shape (batch_size, *sample_shape, prediction_length)
        """

        numel_batch = self.numel_batch
        prediction_length = self.prediction_length

        num_samples_per_batch = prod(sample_shape)
        num_samples = num_samples_per_batch * numel_batch

        hidden_state_repeat = self.hidden_state.repeat_interleave(
            repeats=num_samples_per_batch, dim=0
        )

        alpha = torch.rand(
            (num_samples, prediction_length),
            dtype=self.hidden_state.dtype,
            device=self.hidden_state.device,
            layout=self.hidden_state.layout,
        )

        return self.quantile(alpha, hidden_state_repeat).reshape(
            (numel_batch,) + sample_shape + (prediction_length,)
        )

    def quantile(
        self, alpha: torch.Tensor, hidden_state: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """
        Generates the predicted paths associated with the quantile levels
        alpha.

        Parameters
        ----------
        alpha
            quantile levels,
            shape = (batch_shape, prediction_length)
        hidden_state
            hidden_state, shape = (batch_shape, hidden_size)

        Returns
        -------
        results
            predicted paths of shape = (batch_shape, prediction_length)
        """

        if hidden_state is None:
            hidden_state = self.hidden_state

        normal_quantile = self.standard_normal.icdf(alpha)

        # In the energy score approach, we directly draw samples from picnn
        # In the MLE (Normalizing flows) approach, we need to invert the picnn
        # (go backward through the flow) to draw samples
        if self.is_energy_score:
            result = self.picnn(normal_quantile, context=hidden_state)
        else:
            result = self.picnn.reverse(normal_quantile, context=hidden_state)

        return result

    @property
    def batch_shape(self) -> torch.Size:
        # last dimension is the hidden state size
        return self.hidden_state.shape[:-1]

    @property
    def event_shape(self) -> torch.Size:
        return torch.Size()

    @property
    def event_dim(self) -> int:
        return 0


class MQF2DistributionOutput(DistributionOutput):
    distr_cls: type = MQF2Distribution

    @validated()
    def __init__(
        self,
        prediction_length: int,
        is_energy_score: bool = True,
        threshold_input: float = 100.0,
        es_num_samples: int = 50,
        beta: float = 1.0,
    ) -> None:
        super().__init__(self, beta=beta)
        # A null args_dim to be called by PtArgProj
        self.args_dim = cast(
            Dict[str, int],
            {"null": 1},
        )

        self.prediction_length = prediction_length
        self.is_energy_score = is_energy_score
        self.threshold_input = threshold_input
        self.es_num_samples = es_num_samples

    @classmethod
    def domain_map(  # type: ignore
        cls,
        hidden_state: torch.Tensor,
    ) -> Tuple:
        # A null function to be called by ArgProj
        return ()

    def distribution(  # type: ignore
        self,
        picnn: torch.nn.Module,
        hidden_state: torch.Tensor,
        scale: Optional[torch.Tensor] = None,
    ) -> MQF2Distribution:
        distr = self.distr_cls(
            picnn,
            hidden_state,
            prediction_length=self.prediction_length,
            threshold_input=self.threshold_input,
            es_num_samples=self.es_num_samples,
            is_energy_score=self.is_energy_score,
            beta=self.beta,
        )

        if scale is None:
            return distr
        else:
            return TransformedMQF2Distribution(
                distr,
                [AffineTransform(loc=0.0, scale=scale)],
                is_energy_score=self.is_energy_score,
            )

    @property
    def event_shape(self) -> Tuple:
        return ()


class TransformedMQF2Distribution(TransformedDistribution):
    @validated()
    def __init__(
        self,
        base_distribution: MQF2Distribution,
        transforms: List[AffineTransform],
        is_energy_score: bool = True,
        validate_args: bool = False,
    ) -> None:
        super().__init__(
            base_distribution, transforms, validate_args=validate_args
        )
        self.is_energy_score = is_energy_score

    def scale_input(
        self, y: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        # Auxiliary function to scale the observations
        z = y
        scale = 1.0
        for t in self.transforms[::-1]:
            assert isinstance(t, AffineTransform), "Not an AffineTransform"
            z = t._inverse(y)
            scale *= t.scale

        assert isinstance(scale, torch.Tensor)

        return z, scale

    def repeat_scale(self, scale: torch.Tensor) -> torch.Tensor:
        return scale.squeeze(-1).repeat_interleave(
            self.base_dist.context_length, 0
        )

    def log_prob(self, y: torch.Tensor) -> torch.Tensor:
        prediction_length = self.base_dist.prediction_length

        z, scale = self.scale_input(y)
        p = self.base_dist.log_prob(z)

        repeated_scale = self.repeat_scale(scale)

        # the log scale term can be omitted
        # in optimization because it is a constant
        # prediction_length is the dimension of each sample
        return p - prediction_length * torch.log(repeated_scale)

    def energy_score(self, y: torch.Tensor) -> torch.Tensor:
        beta = self.base_dist.beta

        z, scale = self.scale_input(y)
        loss = self.base_dist.energy_score(z)

        repeated_scale = self.repeat_scale(scale)

        return loss * (repeated_scale**beta)

    def loss(self, z: torch.Tensor) -> torch.Tensor:
        if self.is_energy_score:
            return self.energy_score(z)
        else:
            return -self.log_prob(z)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import tempfile
from functools import partial
from itertools import islice
from pathlib import Path

import pytest
import pandas as pd
import numpy as np
from lightning import seed_everything

from gluonts_customized.dataset.repository import get_dataset
from gluonts_customized.model.predictor import Predictor
from gluonts_customized.torch.distributions import QuantileOutput, StudentTOutput
from gluonts_customized.torch.model.deepar import DeepAREstimator
from gluonts_customized.torch.model.deep_npts import (
    DeepNPTSEstimator,
    DeepNPTSNetworkDiscrete,
    DeepNPTSNetworkSmooth,
)
from gluonts_customized.torch.model.forecast import DistributionForecast
from gluonts_customized.torch.model.mqf2 import MQF2MultiHorizonEstimator
from gluonts_customized.torch.model.simple_feedforward import SimpleFeedForwardEstimator
from gluonts_customized.torch.model.d_linear import DLinearEstimator
from gluonts_customized.torch.model.patch_tst import PatchTSTEstimator
from gluonts_customized.torch.model.tide import TiDEEstimator
from gluonts_customized.torch.model.lag_tst import LagTSTEstimator
from gluonts_customized.torch.model.tft import TemporalFusionTransformerEstimator
from gluonts_customized.torch.model.wavenet import WaveNetEstimator
from gluonts_customized.torch.distributions import ImplicitQuantileNetworkOutput


@pytest.mark.parametrize(
    "estimator_constructor",
    [
        lambda dataset: DeepAREstimator(
            freq=dataset.metadata.freq,
            prediction_length=dataset.metadata.prediction_length,
            distr_output=StudentTOutput(beta=0.1),
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
            scaling=False,
        ),
        lambda dataset: DeepAREstimator(
            freq=dataset.metadata.freq,
            prediction_length=dataset.metadata.prediction_length,
            context_length=1,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
            scaling=False,
        ),
        lambda dataset: MQF2MultiHorizonEstimator(
            freq=dataset.metadata.freq,
            prediction_length=dataset.metadata.prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: SimpleFeedForwardEstimator(
            prediction_length=dataset.metadata.prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: SimpleFeedForwardEstimator(
            prediction_length=dataset.metadata.prediction_length,
            distr_output=QuantileOutput(quantiles=[0.1, 0.6, 0.85]),
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: DLinearEstimator(
            prediction_length=dataset.metadata.prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: DLinearEstimator(
            prediction_length=dataset.metadata.prediction_length,
            distr_output=QuantileOutput(quantiles=[0.1, 0.6, 0.85]),
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: LagTSTEstimator(
            prediction_length=dataset.metadata.prediction_length,
            freq=dataset.metadata.freq,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: TemporalFusionTransformerEstimator(
            freq=dataset.metadata.freq,
            prediction_length=dataset.metadata.prediction_length,
            distr_output=StudentTOutput(),
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: TemporalFusionTransformerEstimator(
            freq=dataset.metadata.freq,
            prediction_length=dataset.metadata.prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: DeepNPTSEstimator(
            freq=dataset.metadata.freq,
            prediction_length=dataset.metadata.prediction_length,
            context_length=2 * dataset.metadata.prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            epochs=2,
        ),
        lambda dataset: PatchTSTEstimator(
            prediction_length=dataset.metadata.prediction_length,
            patch_len=16,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: PatchTSTEstimator(
            prediction_length=dataset.metadata.prediction_length,
            distr_output=QuantileOutput(quantiles=[0.1, 0.6, 0.85]),
            patch_len=16,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: TiDEEstimator(
            freq=dataset.metadata.freq,
            prediction_length=dataset.metadata.prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: TiDEEstimator(
            freq=dataset.metadata.freq,
            prediction_length=dataset.metadata.prediction_length,
            distr_output=QuantileOutput(quantiles=[0.1, 0.6, 0.85]),
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda dataset: WaveNetEstimator(
            freq=dataset.metadata.freq,
            prediction_length=dataset.metadata.prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
    ],
)
@pytest.mark.parametrize("use_validation_data", [False, True])
def test_estimator_constant_dataset(
    estimator_constructor, use_validation_data: bool
):
    constant = get_dataset("constant")

    estimator = estimator_constructor(constant)

    if use_validation_data:
        predictor = estimator.train(
            training_data=constant.train,
            validation_data=constant.train,
        )
    else:
        predictor = estimator.train(
            training_data=constant.train,
        )

    with tempfile.TemporaryDirectory() as td:
        predictor.serialize(Path(td))
        predictor_copy = Predictor.deserialize(Path(td))

    assert predictor == predictor_copy

    forecasts = predictor_copy.predict(constant.test)

    for f in islice(forecasts, 5):
        if isinstance(f, DistributionForecast):
            f = f.to_sample_forecast()
        f.mean


@pytest.mark.parametrize(
    "estimator_constructor",
    [
        lambda freq, prediction_length: DeepAREstimator(
            freq=freq,
            prediction_length=prediction_length,
            distr_output=StudentTOutput(beta=0.1),
            batch_size=4,
            num_batches_per_epoch=3,
            num_feat_dynamic_real=3,
            num_feat_static_real=1,
            num_feat_static_cat=2,
            cardinality=[2, 2],
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda freq, prediction_length: DeepAREstimator(
            freq=freq,
            prediction_length=prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            num_feat_dynamic_real=3,
            num_feat_static_real=1,
            num_feat_static_cat=2,
            cardinality=[2, 2],
            trainer_kwargs=dict(max_epochs=2),
            distr_output=ImplicitQuantileNetworkOutput(),
        ),
        lambda freq, prediction_length: TiDEEstimator(
            freq=freq,
            prediction_length=prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            num_feat_dynamic_real=3,
            num_feat_static_real=1,
            num_feat_static_cat=2,
            cardinality=[2, 2],
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda freq, prediction_length: MQF2MultiHorizonEstimator(
            freq=freq,
            prediction_length=prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            num_feat_dynamic_real=3,
            num_feat_static_real=1,
            num_feat_static_cat=2,
            cardinality=[2, 2],
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda freq, prediction_length: TemporalFusionTransformerEstimator(
            freq=freq,
            prediction_length=prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            dynamic_dims=[3],
            static_dims=[1],
            static_cardinalities=[2, 2],
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda freq, prediction_length: DeepNPTSEstimator(
            freq=freq,
            prediction_length=prediction_length,
            context_length=2 * prediction_length,
            batch_norm=True,
            network_type=partial(DeepNPTSNetworkDiscrete, use_softmax=True),
            use_feat_static_cat=True,
            cardinality=[2, 2],
            num_feat_static_real=1,
            num_feat_dynamic_real=0,
            input_scaling=None,
            dropout_rate=0.0,
            batch_size=4,
            num_batches_per_epoch=3,
            epochs=2,
        ),
        lambda freq, prediction_length: DeepNPTSEstimator(
            freq=freq,
            prediction_length=prediction_length,
            context_length=2 * prediction_length,
            batch_norm=False,
            network_type=partial(DeepNPTSNetworkDiscrete, use_softmax=False),
            use_feat_static_cat=True,
            cardinality=[2, 2],
            num_feat_static_real=1,
            num_feat_dynamic_real=0,
            input_scaling="min_max_scaling",
            dropout_rate=0.0,
            batch_size=4,
            num_batches_per_epoch=3,
            epochs=2,
        ),
        lambda freq, prediction_length: DeepNPTSEstimator(
            freq=freq,
            prediction_length=prediction_length,
            context_length=2 * prediction_length,
            batch_norm=True,
            network_type=DeepNPTSNetworkSmooth,
            use_feat_static_cat=True,
            cardinality=[2, 2],
            num_feat_static_real=1,
            num_feat_dynamic_real=0,
            input_scaling="standard_normal_scaling",
            dropout_rate=0.1,
            batch_size=4,
            num_batches_per_epoch=3,
            epochs=2,
        ),
        lambda freq, prediction_length: PatchTSTEstimator(
            prediction_length=prediction_length,
            context_length=2 * prediction_length,
            num_feat_dynamic_real=3,
            patch_len=16,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda freq, prediction_length: PatchTSTEstimator(
            prediction_length=prediction_length,
            context_length=2 * prediction_length,
            num_feat_dynamic_real=3,
            distr_output=QuantileOutput(quantiles=[0.1, 0.6, 0.85]),
            patch_len=16,
            batch_size=4,
            num_batches_per_epoch=3,
            trainer_kwargs=dict(max_epochs=2),
        ),
        lambda freq, prediction_length: WaveNetEstimator(
            freq=freq,
            prediction_length=prediction_length,
            batch_size=4,
            num_batches_per_epoch=3,
            num_feat_dynamic_real=3,
            num_feat_static_real=1,
            num_feat_static_cat=2,
            cardinality=[2, 2],
            trainer_kwargs=dict(max_epochs=2),
        ),
    ],
)
def test_estimator_with_features(estimator_constructor):
    seed_everything(42)
    freq = "1h"
    prediction_length = 12

    training_dataset = [
        {
            "start": pd.Period("2021-01-01 00:00:00", freq=freq),
            "target": np.ones(200, dtype=np.float32),
            "feat_static_cat": np.array([0, 1], dtype=np.float32),
            "feat_static_real": np.array([42.0], dtype=np.float32),
            "feat_dynamic_real": np.ones((3, 200), dtype=np.float32),
            "__unused__": np.ones(3, dtype=np.float32),
        },
        {
            "start": pd.Period("2021-02-01 00:00:00", freq=freq),
            "target": np.ones(100, dtype=np.float32),
            "feat_static_cat": np.array([1, 0], dtype=np.float32),
            "feat_static_real": np.array([1.0], dtype=np.float32),
            "feat_dynamic_real": np.ones((3, 100), dtype=np.float32),
            "__unused__": np.ones(5, dtype=np.float32),
        },
    ]

    prediction_dataset = [
        {
            "start": pd.Period("2021-01-01 00:00:00", freq=freq),
            "target": np.ones(200, dtype=np.float32),
            "feat_static_cat": np.array([0, 1], dtype=np.float32),
            "feat_static_real": np.array([42.0], dtype=np.float32),
            "feat_dynamic_real": np.ones(
                (3, 200 + prediction_length), dtype=np.float32
            ),
            "__unused__": np.ones(3, dtype=np.float32),
        },
        {
            "start": pd.Period("2021-02-01 00:00:00", freq=freq),
            "target": np.ones(100, dtype=np.float32),
            "feat_static_cat": np.array([1, 0], dtype=np.float32),
            "feat_static_real": np.array([1.0], dtype=np.float32),
            "feat_dynamic_real": np.ones(
                (3, 100 + prediction_length), dtype=np.float32
            ),
            "__unused__": np.ones(5, dtype=np.float32),
        },
    ]

    estimator = estimator_constructor(freq, prediction_length)

    predictor = estimator.train(
        training_data=training_dataset,
        validation_data=training_dataset,
    )

    with tempfile.TemporaryDirectory() as td:
        predictor.serialize(Path(td))
        predictor_copy = Predictor.deserialize(Path(td))

    assert predictor == predictor_copy

    forecasts = predictor_copy.predict(prediction_dataset)

    for f in islice(forecasts, 5):
        f.mean
